
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.matrix

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.matrix.generated.*
import ru.dbotthepony.kvector.matrix.generated.matrixConcreteDeterminant
import ru.dbotthepony.kvector.narray.*
import kotlin.math.pow

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Double2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterDouble, matrix2: IMatrixGetterDouble): Double2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsDouble.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Double2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0.0

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum
		}
	}

	return output
}

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Float2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterFloat, matrix2: IMatrixGetterFloat): Float2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsFloat.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Float2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0f

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum
		}
	}

	return output
}

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Int2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterInt, matrix2: IMatrixGetterInt): Int2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsInt.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Int2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum
		}
	}

	return output
}

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Long2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterLong, matrix2: IMatrixGetterLong): Long2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsLong.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Long2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0L

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum
		}
	}

	return output
}

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Short2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterShort, matrix2: IMatrixGetterShort): Short2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsShort.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Short2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum.toShort()
		}
	}

	return output
}

/**
 * Multiplies matrices as long as they are compatible, of any size.
 * Attempts to use concrete implementation first.
 *
 * @throws IllegalArgumentException if matrices are incompatible
 *
 * @return new [Byte2Dimensional] buffer with multiplication result
 */
fun multiplyMatrix(matrix1: IMatrixGetterByte, matrix2: IMatrixGetterByte): Byte2Dimensional {
	require(matrix1.columns == matrix2.rows) { "Matrices are incompatible for multiplication (Matrix 1 has ${matrix1.columns} columns, Matrix 2 has ${matrix2.rows} rows)" }

	val multiplied = MultiplicationsByte.multiplyMatrix(matrix1, matrix2)

	if (multiplied != null) {
		return multiplied
	}

	val resultRows = matrix1.rows
	val resultColumns = matrix2.columns
	val vectorized = matrix1.columns

	val output = Byte2Dimensional(resultColumns, resultRows)

	for (row in 0 until resultRows) {
		for (column in 0 until resultColumns) {
			var sum = 0L

			for (rowColumn in 0 until vectorized) {
				sum += matrix1[row, rowColumn] * matrix2[rowColumn, column]
			}

			output[row, column] = sum.toByte()
		}
	}

	return output
}

/**
 * Transposes structure representing matrix of any size.
 */
fun transposeMatrix(matrix: IMatrixGetterDouble): Double2Dimensional {
	val output = Double2Dimensional(matrix.rows, matrix.columns)

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			output[row, column] = matrix[column, row]
		}
	}

	return output
}

/**
 * Transposes structure representing matrix of any size.
 */
fun transposeMatrix(matrix: IMatrixGetterFloat): Float2Dimensional {
	val output = Float2Dimensional(matrix.rows, matrix.columns)

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			output[row, column] = matrix[column, row]
		}
	}

	return output
}

private fun load(matrix: IMatrixGetterDouble, buffers: Array<Double2Dimensional>): Double2Dimensional {
	val buffer = buffers[matrix.columns - 1]

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			buffer[column, row] = matrix[column, row]
		}
	}

	return buffer
}

private fun load(matrix: IMatrixGetterFloat, buffers: Array<Float2Dimensional>): Float2Dimensional {
	val buffer = buffers[matrix.columns - 1]

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			buffer[column, row] = matrix[column, row]
		}
	}

	return buffer
}

private fun load(matrix: IMatrixGetterInt, buffers: Array<Int2Dimensional>): Int2Dimensional {
	val buffer = buffers[matrix.columns - 1]

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			buffer[column, row] = matrix[column, row]
		}
	}

	return buffer
}

private fun load(matrix: IMatrixGetterLong, buffers: Array<Long2Dimensional>): Long2Dimensional {
	val buffer = buffers[matrix.columns - 1]

	for (column in 0 until matrix.columns) {
		for (row in 0 until matrix.rows) {
			buffer[column, row] = matrix[column, row]
		}
	}

	return buffer
}

/**
 * Constructs new complement matrix from [input] representing square matrix,
 * by taking out [column] and [row], and load result into [buffer].
 *
 * @throws IllegalArgumentException if [column] or [row] is out of bounds,
 * or if [input] does not represent a square matrix
 *
 * @return [buffer]
 */
fun complementMatrix(input: IMatrixGetterDouble, column: Int, row: Int, buffer: Double2Dimensional): Double2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	require(column < input.columns) { "$column < ${input.columns}" }
	require(row < input.rows) { "$row < ${input.rows}" }

	var iRow = 0
	var iColumn = 0

	for (gColumn in 0 until input.columns) {
		if (gColumn == column)
			continue

		for (gRow in 0 until input.rows) {
			if (gRow == row)
				continue

			buffer[iColumn, iRow++] = input[gColumn, gRow]
		}

		iColumn++
		iRow = 0
	}

	return buffer
}

/**
 * Constructs new complement matrix from [input] representing square matrix,
 * by taking out [column] and [row], and load result into [buffer].
 *
 * @throws IllegalArgumentException if [column] or [row] is out of bounds,
 * or if [input] does not represent a square matrix
 *
 * @return [buffer]
 */
fun complementMatrix(input: IMatrixGetterFloat, column: Int, row: Int, buffer: Float2Dimensional): Float2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	require(column < input.columns) { "$column < ${input.columns}" }
	require(row < input.rows) { "$row < ${input.rows}" }

	var iRow = 0
	var iColumn = 0

	for (gColumn in 0 until input.columns) {
		if (gColumn == column)
			continue

		for (gRow in 0 until input.rows) {
			if (gRow == row)
				continue

			buffer[iColumn, iRow++] = input[gColumn, gRow]
		}

		iColumn++
		iRow = 0
	}

	return buffer
}

/**
 * Constructs new complement matrix from [input] representing square matrix,
 * by taking out [column] and [row], and load result into [buffer].
 *
 * @throws IllegalArgumentException if [column] or [row] is out of bounds,
 * or if [input] does not represent a square matrix
 *
 * @return [buffer]
 */
fun complementMatrix(input: IMatrixGetterInt, column: Int, row: Int, buffer: Int2Dimensional): Int2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	require(column < input.columns) { "$column < ${input.columns}" }
	require(row < input.rows) { "$row < ${input.rows}" }

	var iRow = 0
	var iColumn = 0

	for (gColumn in 0 until input.columns) {
		if (gColumn == column)
			continue

		for (gRow in 0 until input.rows) {
			if (gRow == row)
				continue

			buffer[iColumn, iRow++] = input[gColumn, gRow]
		}

		iColumn++
		iRow = 0
	}

	return buffer
}

/**
 * Constructs new complement matrix from [input] representing square matrix,
 * by taking out [column] and [row], and load result into [buffer].
 *
 * @throws IllegalArgumentException if [column] or [row] is out of bounds,
 * or if [input] does not represent a square matrix
 *
 * @return [buffer]
 */
fun complementMatrix(input: IMatrixGetterLong, column: Int, row: Int, buffer: Long2Dimensional): Long2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	require(column < input.columns) { "$column < ${input.columns}" }
	require(row < input.rows) { "$row < ${input.rows}" }

	var iRow = 0
	var iColumn = 0

	for (gColumn in 0 until input.columns) {
		if (gColumn == column)
			continue

		for (gRow in 0 until input.rows) {
			if (gRow == row)
				continue

			buffer[iColumn, iRow++] = input[gColumn, gRow]
		}

		iColumn++
		iRow = 0
	}

	return buffer
}

/**
 * Computes cofactor matrix of specified square matrix [input],
 * returning result as [Double2Dimensional]
 *
 * This operation requires n * n determinant calculations of matrix,
 * so do cache result of this function in performance critical context.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix
 */
fun cofactorMatrix(input: IMatrixGetterDouble): Double2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	if (input.rows == 1) {
		return Double2Dimensional(1, 1).also { it[0, 0] = input[0, 0] }
	} else if (input.rows == 2) {
		return Double2Dimensional(2, 2).also {
			it[0, 0] = input[1, 1]
			it[1, 0] = -input[1, 0]
			it[0, 1] = -input[0, 1]
			it[1, 1] = input[0, 0]
		}
	}

	val result = Double2Dimensional(input.columns, input.rows)
	val buffers = Array(input.rows) { Double2Dimensional(it + 1, it + 1) }

	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			val minor = matrixDeterminant(complementMatrix(input, column, row, buffers[input.columns - 2]), input.columns - 1, buffers)
			val cofactor = (-1.0).pow(row + column) * minor
			result[column, row] = cofactor
		}
	}

	return result
}

/**
 * Computes cofactor matrix of specified square matrix [input],
 * returning result as [Float2Dimensional]
 *
 * This operation requires n * n determinant calculations of matrix,
 * so do cache result of this function in performance critical context.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix
 */
fun cofactorMatrix(input: IMatrixGetterFloat): Float2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	if (input.rows == 1) {
		return Float2Dimensional(1, 1).also { it[0, 0] = input[0, 0] }
	} else if (input.rows == 2) {
		return Float2Dimensional(2, 2).also {
			it[0, 0] = input[1, 1]
			it[1, 0] = -input[1, 0]
			it[0, 1] = -input[0, 1]
			it[1, 1] = input[0, 0]
		}
	}

	val result = Float2Dimensional(input.columns, input.rows)
	val buffers = Array(input.rows) { Float2Dimensional(it + 1, it + 1) }

	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			val minor = matrixDeterminant(complementMatrix(input, column, row, buffers[input.columns - 2]), input.columns - 1, buffers)
			val cofactor = (-1f).pow(row + column) * minor
			result[column, row] = cofactor
		}
	}

	return result
}

/**
 * Computes adjugate matrix of specified square matrix [input],
 * returning result as [Double2Dimensional],
 *
 * [https://en.wikipedia.org/wiki/Adjugate_matrix]
 *
 * This operation requires n * n determinant calculations of matrix,
 * so do cache result of this function in performance critical context.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix
 */
fun adjugateMatrix(input: IMatrixGetterDouble): Double2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	return transposeMatrix(cofactorMatrix(input))
}

/**
 * Computes adjugate matrix of specified square matrix [input],
 * returning result as [Float2Dimensional],
 *
 * [https://en.wikipedia.org/wiki/Adjugate_matrix]
 *
 * This operation requires n * n determinant calculations of matrix,
 * so do cache result of this function in performance critical context.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix
 */
fun adjugateMatrix(input: IMatrixGetterFloat): Float2Dimensional {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }
	return transposeMatrix(cofactorMatrix(input))
}

/**
 * Computes inverse matrix of specified square matrix [input],
 * returning result as [Double2Dimensional],
 *
 * This operation requires 1 + n * n determinant calculations of original matrix,
 * so do cache result of this function in performance critical context.
 *
 * If determinant is zero, null is returned.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix, or null if determinant is zero.
 */
fun inverseMatrix(input: IMatrixGetterDouble): Double2Dimensional? {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	var determinant = matrixDeterminant(input)

	if (determinant == 0.0)
		return null

	determinant = 1.0 / determinant
	val adjugate = adjugateMatrix(input)

	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			adjugate[column, row] *= determinant
		}
	}

	return adjugate
}

/**
 * Computes inverse matrix of specified square matrix [input],
 * returning result as [Float2Dimensional],
 *
 * This operation requires 1 + n * n determinant calculations of original matrix,
 * so do cache result of this function in performance critical context.
 *
 * If determinant is zero, null is returned.
 *
 * @throws IllegalArgumentException if provided [input] is not a square matrix
 * @return 2d array containing result matrix, or null if determinant is zero.
 */
fun inverseMatrix(input: IMatrixGetterFloat): Float2Dimensional? {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	var determinant = matrixDeterminant(input)

	if (determinant == 0f)
		return null

	determinant = 1f / determinant
	val adjugate = adjugateMatrix(input)

	for (column in 0 until input.columns) {
		for (row in 0 until input.rows) {
			adjugate[column, row] *= determinant
		}
	}

	return adjugate
}

private fun matrixDeterminant(input: Double2Dimensional, size: Int, buffers: Array<Double2Dimensional>): Double {
	if (input.rows == 1)
		return input[0, 0]

	val concrete = matrixConcreteDeterminant(input)

	if (concrete != null) {
		return concrete
	}

	var result = 0.0

	for (column in 0 until size) {
		val minor = matrixDeterminant(complementMatrix(input, column, 0, buffers[input.columns - 2]), size - 1, buffers)
		val cofactor = (-1.0).pow(2 + column) * minor
		result += input[column, 0] * cofactor
	}

	return result
}

/**
 * Recursive determinant finder of specified structure representing square matrix.
 *
 * @throws IllegalArgumentException if [input] structure is not square matrix
 */
fun matrixDeterminant(input: IMatrixGetterDouble): Double {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	val concrete = matrixConcreteDeterminant(input)

	if (concrete != null) {
		return concrete
	}

	val buffers = Array(input.rows) { Double2Dimensional(it + 1, it + 1) }
	return matrixDeterminant(load(input, buffers), input.columns, buffers)
}

private fun matrixDeterminant(input: Float2Dimensional, size: Int, buffers: Array<Float2Dimensional>): Float {
	val concrete = matrixConcreteDeterminant(input)

	if (concrete != null) {
		return concrete
	}

	var result = 0f

	for (column in 0 until size) {
		val minor = matrixDeterminant(complementMatrix(input, column, 0, buffers[input.columns - 2]), size - 1, buffers)
		val cofactor = (-1f).pow(2 + column) * minor
		result += input[column, 0] * cofactor
	}

	return result
}

/**
 * Recursive determinant finder of specified structure representing square matrix.
 *
 * @throws IllegalArgumentException if [input] structure is not square matrix
 */
fun matrixDeterminant(input: IMatrixGetterFloat): Float {
	require(input.rows == input.columns) { "Provided matrix is not square matrix (${input.columns} columns, ${input.rows} rows)" }

	val concrete = matrixConcreteDeterminant(input)

	if (concrete != null) {
		return concrete
	}

	val buffers = Array(input.rows) { Float2Dimensional(it + 1, it + 1) }
	return matrixDeterminant(load(input, buffers), input.columns, buffers).toFloat()
}

