
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.vector.ndouble

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.vector.nfloat.Vector2f
import kotlin.math.absoluteValue
import kotlin.math.pow
import kotlin.math.sqrt

/**
 * 2D Vector, representing two-dimensional coordinates as [Double]s
 */
open class Vector2d(
	open val x: Double = 0.0,
	open val y: Double = 0.0
) : AbstractVector<Vector2d>(), IFractionalVector<Vector2d>, IDoubleVector<Vector2d>, IStruct2d, IMatrixGetterDouble {
	final override fun component1() = x
	final override fun component2() = y

	constructor(input: IStruct2d) : this(input.component1(), input.component2())

	final override fun get(column: Int, row: Int): Double {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	final override val rows: Int = 2
	final override val lengthSquared: Double
		get() = x * x + y * y

	final override val isFinite: Boolean
		get() = x.isFinite() && y.isFinite()
	final override val isNaN: Boolean
		get() = x.isNaN() || y.isNaN()

	open val r by this::x
	open val g by this::y

	open val s by this::x
	open val t by this::y

	override fun plus(other: Vector2d): Vector2d {
		return Vector2d(x + other.x, y + other.y)
	}

	override fun minus(other: Vector2d): Vector2d {
		return Vector2d(x - other.x, y - other.y)
	}

	override fun times(other: Vector2d): Vector2d {
		return Vector2d(x * other.x, y * other.y)
	}

	override fun div(other: Vector2d): Vector2d {
		return Vector2d(x / other.x, y / other.y)
	}

	override val absoluteValue: Vector2d
		get() = Vector2d(x.absoluteValue, y.absoluteValue)

	override fun coerceAtMost(maximal: Vector2d): Vector2d {
		return Vector2d(x.coerceAtMost(maximal.x), y.coerceAtMost(maximal.y))
	}

	override fun coerceAtLeast(minimal: Vector2d): Vector2d {
		return Vector2d(x.coerceAtLeast(minimal.x), y.coerceAtLeast(minimal.y))
	}

	override fun clamp(minimal: Vector2d, maximal: Vector2d): Vector2d {
		return Vector2d(x.coerceAtLeast(minimal.x).coerceAtMost(maximal.x), y.coerceAtLeast(minimal.y).coerceAtMost(maximal.y))
	}

	override fun unaryMinus(): Vector2d {
		return Vector2d(-x, -y)
	}

	override fun distanceSquared(other: Vector2d): Double {
		val x = x - other.x
		val y = y - other.y

		return x * x + y * y
	}

	override val normalized: Vector2d
		get() {
			val len = length

			if (len == 0.0) {
				return ZERO
			}

			return Vector2d(x / len, y / len)
		}

	override fun times(other: Double): Vector2d {
		return Vector2d(x * other, y * other)
	}

	override fun div(other: Double): Vector2d {
		return Vector2d(x / other, y / other)
	}

	fun times(other: Float): Vector2d {
		return Vector2d(x * other, y * other)
	}

	fun div(other: Float): Vector2d {
		return Vector2d(x / other, y / other)
	}

	fun times(other: Int): Vector2d {
		return Vector2d(x * other, y * other)
	}

	fun div(other: Int): Vector2d {
		return Vector2d(x / other, y / other)
	}

	/**
	 * Calculates vector vector * vector, returning result as [Double]
	 */
	fun cross(other: Vector2d): Double {
		return x * other.y - y * other.x
	}

	/**
	 * Calculates 2D cross product between scalar and vector
	 *
	 * @return new vector
	 */
	fun cross(other: Double): Vector2d {
		return Vector2d(other * y, -other * x)
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector2d): Double {
		return other.x * x + other.y * y
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector3d): Double {
		return other.x * x + other.y * y
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector4d): Double {
		return other.x * x + other.y * y
	}

	/**
	 * Casts components to [Float] and returns new vector as result
	 */
	open fun toFloatVector(): Vector2f {
		return Vector2f(x.toFloat(), y.toFloat())
	}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as Vector2d

		if (x != other.x) return false
		if (y != other.y) return false

		return true
	}

	override fun toString(): String {
		return "[$x $y]"
	}

	override fun hashCode(): Int {
		return 31 * x.hashCode() + y.hashCode()
	}

	inline val xx get() = Vector2d(x, x)
	inline val xy get() = Vector2d(x, y)
	inline val yx get() = Vector2d(y, x)
	inline val yy get() = Vector2d(y, y)
	inline val xxx get() = Vector3d(x, x, x)
	inline val xxy get() = Vector3d(x, x, y)
	inline val xyx get() = Vector3d(x, y, x)
	inline val xyy get() = Vector3d(x, y, y)
	inline val yxx get() = Vector3d(y, x, x)
	inline val yxy get() = Vector3d(y, x, y)
	inline val yyx get() = Vector3d(y, y, x)
	inline val yyy get() = Vector3d(y, y, y)
	inline val xxxx get() = Vector4d(x, x, x, x)
	inline val xxxy get() = Vector4d(x, x, x, y)
	inline val xxyx get() = Vector4d(x, x, y, x)
	inline val xxyy get() = Vector4d(x, x, y, y)
	inline val xyxx get() = Vector4d(x, y, x, x)
	inline val xyxy get() = Vector4d(x, y, x, y)
	inline val xyyx get() = Vector4d(x, y, y, x)
	inline val xyyy get() = Vector4d(x, y, y, y)
	inline val yxxx get() = Vector4d(y, x, x, x)
	inline val yxxy get() = Vector4d(y, x, x, y)
	inline val yxyx get() = Vector4d(y, x, y, x)
	inline val yxyy get() = Vector4d(y, x, y, y)
	inline val yyxx get() = Vector4d(y, y, x, x)
	inline val yyxy get() = Vector4d(y, y, x, y)
	inline val yyyx get() = Vector4d(y, y, y, x)
	inline val yyyy get() = Vector4d(y, y, y, y)
	inline val rr get() = Vector2d(r, r)
	inline val rg get() = Vector2d(r, g)
	inline val gr get() = Vector2d(g, r)
	inline val gg get() = Vector2d(g, g)
	inline val rrr get() = Vector3d(r, r, r)
	inline val rrg get() = Vector3d(r, r, g)
	inline val rgr get() = Vector3d(r, g, r)
	inline val rgg get() = Vector3d(r, g, g)
	inline val grr get() = Vector3d(g, r, r)
	inline val grg get() = Vector3d(g, r, g)
	inline val ggr get() = Vector3d(g, g, r)
	inline val ggg get() = Vector3d(g, g, g)
	inline val rrrr get() = Vector4d(r, r, r, r)
	inline val rrrg get() = Vector4d(r, r, r, g)
	inline val rrgr get() = Vector4d(r, r, g, r)
	inline val rrgg get() = Vector4d(r, r, g, g)
	inline val rgrr get() = Vector4d(r, g, r, r)
	inline val rgrg get() = Vector4d(r, g, r, g)
	inline val rggr get() = Vector4d(r, g, g, r)
	inline val rggg get() = Vector4d(r, g, g, g)
	inline val grrr get() = Vector4d(g, r, r, r)
	inline val grrg get() = Vector4d(g, r, r, g)
	inline val grgr get() = Vector4d(g, r, g, r)
	inline val grgg get() = Vector4d(g, r, g, g)
	inline val ggrr get() = Vector4d(g, g, r, r)
	inline val ggrg get() = Vector4d(g, g, r, g)
	inline val gggr get() = Vector4d(g, g, g, r)
	inline val gggg get() = Vector4d(g, g, g, g)
	inline val ss get() = Vector2d(s, s)
	inline val st get() = Vector2d(s, t)
	inline val ts get() = Vector2d(t, s)
	inline val tt get() = Vector2d(t, t)
	inline val sss get() = Vector3d(s, s, s)
	inline val sst get() = Vector3d(s, s, t)
	inline val sts get() = Vector3d(s, t, s)
	inline val stt get() = Vector3d(s, t, t)
	inline val tss get() = Vector3d(t, s, s)
	inline val tst get() = Vector3d(t, s, t)
	inline val tts get() = Vector3d(t, t, s)
	inline val ttt get() = Vector3d(t, t, t)
	inline val ssss get() = Vector4d(s, s, s, s)
	inline val ssst get() = Vector4d(s, s, s, t)
	inline val ssts get() = Vector4d(s, s, t, s)
	inline val sstt get() = Vector4d(s, s, t, t)
	inline val stss get() = Vector4d(s, t, s, s)
	inline val stst get() = Vector4d(s, t, s, t)
	inline val stts get() = Vector4d(s, t, t, s)
	inline val sttt get() = Vector4d(s, t, t, t)
	inline val tsss get() = Vector4d(t, s, s, s)
	inline val tsst get() = Vector4d(t, s, s, t)
	inline val tsts get() = Vector4d(t, s, t, s)
	inline val tstt get() = Vector4d(t, s, t, t)
	inline val ttss get() = Vector4d(t, t, s, s)
	inline val ttst get() = Vector4d(t, t, s, t)
	inline val ttts get() = Vector4d(t, t, t, s)
	inline val tttt get() = Vector4d(t, t, t, t)

	companion object {
		@JvmField val ZERO = Vector2d()
		@JvmField val POSITIVE_X = Vector2d(x = 1.0)
		@JvmField val NEGATIVE_X = Vector2d(x = -1.0)
		@JvmField val POSITIVE_Y = Vector2d(y = 1.0)
		@JvmField val NEGATIVE_Y = Vector2d(y = -1.0)
	}
}

fun Double.div(other: Vector2d): Vector2d {
	return Vector2d(this / other.x, this / other.y)
}

fun Double.times(other: Vector2d): Vector2d {
	return Vector2d(this * other.x, this * other.y)
}

/**
 * Calculates 2D cross product between scalar and vector
 *
 * @return new vector
 */
fun Double.cross(other: Vector2d): Vector2d {
	return Vector2d(-this * other.y, this * other.x)
}

/**
 * Mutable 2D Vector, representing two-dimensional coordinates as [Double]s
 *
 * It can be safely passed around to methods which expect immutable [Vector2d], **AND** do not store it,
 * as mutable operations are separated from immutable ones
 */
open class MutableVector2d(
	override var x: Double = 0.0,
	override var y: Double = 0.0
) : Vector2d(x, y), IMutableDoubleVector<MutableVector2d>, IMutableFractionalVector<MutableVector2d>, IMutableVector<MutableVector2d, Vector2d>, IMatrixSetterDouble {
	constructor(input: IStruct2d) : this(input.component1(), input.component2())

	override fun set(column: Int, row: Int, value: Double) {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		when (row) {
			0 -> x = value
			1 -> y = value
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	override var r by this::x
	override val g by this::y

	override val s by this::x
	override val t by this::y

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as MutableVector2d

		if (x != other.x) return false
		if (y != other.y) return false

		return true
	}

	override fun plusMut(other: Vector2d): MutableVector2d {
		x += other.x
		y += other.y
		return this
	}

	override fun minusMut(other: Vector2d): MutableVector2d {
		x -= other.x
		y -= other.y
		return this
	}

	override fun timesMut(other: Vector2d): MutableVector2d {
		x *= other.x
		y *= other.y
		return this
	}

	override fun divMut(other: Vector2d): MutableVector2d {
		x /= other.x
		y /= other.y
		return this
	}

	override fun normalize(): Double {
		val len = length

		if (len == 0.0) {
			return len
		}

		x /= len
		y /= len

		return len
	}

	override fun timesMut(other: Double): MutableVector2d {
		x *= other
		y *= other
		return this
	}

	override fun divMut(other: Double): MutableVector2d {
		x /= other
		y /= other
		return this
	}

	/**
	 * Calculates 2D cross product between scalar and vector,
	 * writing result into this vector
	 *
	 * @return this vector
	 */
	fun crossMut(other: Double): Vector2d {
		val x = x
		val y = y
		this.x = other * y
		this.y = -other * x
		return this
	}
}
