
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.vector.nfloat

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import kotlin.math.absoluteValue

/**
 * 2D Vector, representing two-dimensional coordinates as [Float]s
 */
open class Vector2f(
	open val x: Float = 0f,
	open val y: Float = 0f
) : AbstractVector<Vector2f>(), IFractionalVector<Vector2f>, IFloatVector<Vector2f>, IStruct2f, IMatrixGetterFloat {
	final override fun component1() = x
	final override fun component2() = y

	constructor(input: IStruct2f) : this(input.component1(), input.component2())

	final override fun get(column: Int, row: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	final override val rows: Int = 2
	final override val lengthSquared: Double
		get() = x.toDouble() * x.toDouble() + y.toDouble() * y.toDouble()

	final override val isFinite: Boolean
		get() = x.isFinite() && y.isFinite()
	final override val isNaN: Boolean
		get() = x.isNaN() || y.isNaN()

	open val r by this::x
	open val g by this::y

	open val s by this::x
	open val t by this::y

	override fun plus(other: Vector2f): Vector2f {
		return Vector2f(
			x + other.x,
			y + other.y,
		)
	}

	override fun minus(other: Vector2f): Vector2f {
		return Vector2f(
			x - other.x,
			y - other.y,
		)
	}

	override fun times(other: Vector2f): Vector2f {
		return Vector2f(
			x * other.x,
			y * other.y,
		)
	}

	override fun div(other: Vector2f): Vector2f {
		return Vector2f(
			x / other.x,
			y / other.y,
		)
	}

	override val absoluteValue: Vector2f
		get() = Vector2f(x.absoluteValue, y.absoluteValue)

	override fun coerceAtMost(maximal: Vector2f): Vector2f {
		return Vector2f(
			x.coerceAtMost(maximal.x),
			y.coerceAtMost(maximal.y),
		)
	}

	override fun coerceAtLeast(minimal: Vector2f): Vector2f {
		return Vector2f(
			x.coerceAtLeast(minimal.x),
			y.coerceAtLeast(minimal.y),
		)
	}

	override fun clamp(minimal: Vector2f, maximal: Vector2f): Vector2f {
		return Vector2f(
			x.coerceAtLeast(minimal.x).coerceAtMost(maximal.x),
			y.coerceAtLeast(minimal.y).coerceAtMost(maximal.y),
		)
	}

	override fun unaryMinus(): Vector2f {
		return Vector2f(-x, -y)
	}

	override fun distanceSquared(other: Vector2f): Double {
		val x = x.toDouble() - other.x.toDouble()
		val y = y.toDouble() - other.y.toDouble()

		return x * x + y * y
	}

	override val normalized: Vector2f
		get() {
			val len = length

			if (len == 0.0) {
				return ZERO
			}

			return Vector2f((x / len).toFloat(), (y / len).toFloat())
		}

	override fun times(other: Float): Vector2f {
		return Vector2f(
			x * other,
			y * other
		)
	}

	override fun div(other: Float): Vector2f {
		return Vector2f(
			x / other,
			y / other
		)
	}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as Vector2f

		if (x != other.x) return false
		if (y != other.y) return false

		return true
	}

	override fun hashCode(): Int {
		var result = x.hashCode()
		result = 31 * result + y.hashCode()
		return result
	}

	override fun toString(): String {
		return "[${x}f ${y}f]"
	}

	/**
	 * Calculates vector vector * vector, returning result as [Double]
	 */
	fun cross(other: Vector2f): Double {
		return x.toDouble() * other.y.toDouble() - y.toDouble() * other.x.toDouble()
	}

	/**
	 * Calculates 2D cross product between scalar and vector
	 *
	 * @return new vector
	 */
	fun cross(other: Float): Vector2f {
		return Vector2f(other * y, -other * x)
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector2f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble()
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector3f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble()
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector4f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble()
	}

	/**
	 * Casts components to [Double] and returns new vector as result
	 */
	open fun toDoubleVector(): Vector2d {
		return Vector2d(x.toDouble(), y.toDouble())
	}

	inline val xx get() = Vector2f(x, x)
	inline val xy get() = Vector2f(x, y)
	inline val yx get() = Vector2f(y, x)
	inline val yy get() = Vector2f(y, y)
	inline val xxx get() = Vector3f(x, x, x)
	inline val xxy get() = Vector3f(x, x, y)
	inline val xyx get() = Vector3f(x, y, x)
	inline val xyy get() = Vector3f(x, y, y)
	inline val yxx get() = Vector3f(y, x, x)
	inline val yxy get() = Vector3f(y, x, y)
	inline val yyx get() = Vector3f(y, y, x)
	inline val yyy get() = Vector3f(y, y, y)
	inline val xxxx get() = Vector4f(x, x, x, x)
	inline val xxxy get() = Vector4f(x, x, x, y)
	inline val xxyx get() = Vector4f(x, x, y, x)
	inline val xxyy get() = Vector4f(x, x, y, y)
	inline val xyxx get() = Vector4f(x, y, x, x)
	inline val xyxy get() = Vector4f(x, y, x, y)
	inline val xyyx get() = Vector4f(x, y, y, x)
	inline val xyyy get() = Vector4f(x, y, y, y)
	inline val yxxx get() = Vector4f(y, x, x, x)
	inline val yxxy get() = Vector4f(y, x, x, y)
	inline val yxyx get() = Vector4f(y, x, y, x)
	inline val yxyy get() = Vector4f(y, x, y, y)
	inline val yyxx get() = Vector4f(y, y, x, x)
	inline val yyxy get() = Vector4f(y, y, x, y)
	inline val yyyx get() = Vector4f(y, y, y, x)
	inline val yyyy get() = Vector4f(y, y, y, y)
	inline val rr get() = Vector2f(r, r)
	inline val rg get() = Vector2f(r, g)
	inline val gr get() = Vector2f(g, r)
	inline val gg get() = Vector2f(g, g)
	inline val rrr get() = Vector3f(r, r, r)
	inline val rrg get() = Vector3f(r, r, g)
	inline val rgr get() = Vector3f(r, g, r)
	inline val rgg get() = Vector3f(r, g, g)
	inline val grr get() = Vector3f(g, r, r)
	inline val grg get() = Vector3f(g, r, g)
	inline val ggr get() = Vector3f(g, g, r)
	inline val ggg get() = Vector3f(g, g, g)
	inline val rrrr get() = Vector4f(r, r, r, r)
	inline val rrrg get() = Vector4f(r, r, r, g)
	inline val rrgr get() = Vector4f(r, r, g, r)
	inline val rrgg get() = Vector4f(r, r, g, g)
	inline val rgrr get() = Vector4f(r, g, r, r)
	inline val rgrg get() = Vector4f(r, g, r, g)
	inline val rggr get() = Vector4f(r, g, g, r)
	inline val rggg get() = Vector4f(r, g, g, g)
	inline val grrr get() = Vector4f(g, r, r, r)
	inline val grrg get() = Vector4f(g, r, r, g)
	inline val grgr get() = Vector4f(g, r, g, r)
	inline val grgg get() = Vector4f(g, r, g, g)
	inline val ggrr get() = Vector4f(g, g, r, r)
	inline val ggrg get() = Vector4f(g, g, r, g)
	inline val gggr get() = Vector4f(g, g, g, r)
	inline val gggg get() = Vector4f(g, g, g, g)
	inline val ss get() = Vector2f(s, s)
	inline val st get() = Vector2f(s, t)
	inline val ts get() = Vector2f(t, s)
	inline val tt get() = Vector2f(t, t)
	inline val sss get() = Vector3f(s, s, s)
	inline val sst get() = Vector3f(s, s, t)
	inline val sts get() = Vector3f(s, t, s)
	inline val stt get() = Vector3f(s, t, t)
	inline val tss get() = Vector3f(t, s, s)
	inline val tst get() = Vector3f(t, s, t)
	inline val tts get() = Vector3f(t, t, s)
	inline val ttt get() = Vector3f(t, t, t)
	inline val ssss get() = Vector4f(s, s, s, s)
	inline val ssst get() = Vector4f(s, s, s, t)
	inline val ssts get() = Vector4f(s, s, t, s)
	inline val sstt get() = Vector4f(s, s, t, t)
	inline val stss get() = Vector4f(s, t, s, s)
	inline val stst get() = Vector4f(s, t, s, t)
	inline val stts get() = Vector4f(s, t, t, s)
	inline val sttt get() = Vector4f(s, t, t, t)
	inline val tsss get() = Vector4f(t, s, s, s)
	inline val tsst get() = Vector4f(t, s, s, t)
	inline val tsts get() = Vector4f(t, s, t, s)
	inline val tstt get() = Vector4f(t, s, t, t)
	inline val ttss get() = Vector4f(t, t, s, s)
	inline val ttst get() = Vector4f(t, t, s, t)
	inline val ttts get() = Vector4f(t, t, t, s)
	inline val tttt get() = Vector4f(t, t, t, t)

	companion object {
		@JvmField val ZERO = Vector2f()
		@JvmField val POSITIVE_X = Vector2f(x = 1f)
		@JvmField val NEGATIVE_X = Vector2f(x = -1f)
		@JvmField val POSITIVE_Y = Vector2f(y = 1f)
		@JvmField val NEGATIVE_Y = Vector2f(y = -1f)
	}
}

operator fun Float.times(other: Vector2f): Vector2f {
	return Vector2f(this * other.x, this * other.y)
}

operator fun Float.div(other: Vector2f): Vector2f {
	return Vector2f(this / other.x, this / other.y)
}

/**
 * Mutable 2D Vector, representing two-dimensional coordinates as [Float]s
 *
 * It can be safely passed around to methods which expect immutable [Vector2f], **AND** do not store it,
 * as mutable operations are separated from immutable ones
 */
open class MutableVector2f(
	override var x: Float = 0f,
	override var y: Float = 0f
) : Vector2f(x, y), IMutableFloatVector<MutableVector2f>, IMutableFractionalVector<MutableVector2f>, IMutableVector<MutableVector2f, Vector2f>, IMatrixSetterFloat {
	constructor(input: IStruct2f) : this(input.component1(), input.component2())

	override fun set(column: Int, row: Int, value: Float) {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		when (row) {
			0 -> x = value
			1 -> y = value
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	override var r by this::x
	override var g by this::y

	override var s by this::x
	override var t by this::y

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as MutableVector2f

		if (x != other.x) return false
		if (y != other.y) return false

		return true
	}

	override fun normalize(): Double {
		val len = length

		if (len == 0.0) {
			return len
		}

		x = (x / len).toFloat()
		y = (y / len).toFloat()

		return len
	}

	override fun plusMut(other: Vector2f): MutableVector2f {
		x += other.x
		y += other.y
		return this
	}

	override fun minusMut(other: Vector2f): MutableVector2f {
		x -= other.x
		y -= other.y
		return this
	}

	override fun timesMut(other: Vector2f): MutableVector2f {
		x *= other.x
		y *= other.y
		return this
	}

	override fun divMut(other: Vector2f): MutableVector2f {
		x /= other.x
		y /= other.y
		return this
	}

	override fun timesMut(other: Float): MutableVector2f {
		x *= other
		y *= other
		return this
	}

	override fun divMut(other: Float): MutableVector2f {
		x /= other
		y /= other
		return this
	}
}
