
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.vector.nfloat

import ru.dbotthepony.kvector.api.*
import kotlin.math.absoluteValue

/**
 * 3D Vector, representing three-dimensional coordinates as [Float]s
 */
open class Vector3f(
	open val x: Float = 0f,
	open val y: Float = 0f,
	open val z: Float = 0f,
) : AbstractVector<Vector3f>(), IFractionalVector<Vector3f>, IFloatVector<Vector3f>, IStruct3f, IMatrixGetterFloat {
	final override fun component1() = x
	final override fun component2() = y
	final override fun component3() = z

	override fun get(column: Int, row: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	constructor(input: IStruct2f, z: Float = 0f) : this(input.component1(), input.component2(), z)
	constructor(x: Float, input: IStruct2f) : this(x, input.component1(), input.component2())
	constructor(input: IStruct3f) : this(input.component1(), input.component2(), input.component3())

	final override val rows: Int = 3
	final override val lengthSquared: Double
		get() = x.toDouble() * x.toDouble() + y.toDouble() * y.toDouble() + z.toDouble() * z.toDouble()

	final override val isFinite: Boolean
		get() = x.isFinite() && y.isFinite() && z.isFinite()
	final override val isNaN: Boolean
		get() = x.isNaN() || y.isNaN() || z.isNaN()

	open val r by this::x
	open val g by this::y
	open val b by this::z

	open val s by this::x
	open val t by this::y
	open val p by this::z

	override fun plus(other: Vector3f): Vector3f {
		return Vector3f(
			x + other.x,
			y + other.y,
			z + other.z,
		)
	}

	override fun minus(other: Vector3f): Vector3f {
		return Vector3f(
			x - other.x,
			y - other.y,
			z - other.z,
		)
	}

	override fun times(other: Vector3f): Vector3f {
		return Vector3f(
			x * other.x,
			y * other.y,
			z * other.z,
		)
	}

	override fun div(other: Vector3f): Vector3f {
		return Vector3f(
			x / other.x,
			y / other.y,
			z / other.z,
		)
	}

	override val absoluteValue: Vector3f
		get() = Vector3f(x.absoluteValue, y.absoluteValue, z.absoluteValue)

	override fun coerceAtMost(maximal: Vector3f): Vector3f {
		return Vector3f(
			x.coerceAtMost(maximal.x),
			y.coerceAtMost(maximal.y),
			z.coerceAtMost(maximal.z),
		)
	}

	override fun coerceAtLeast(minimal: Vector3f): Vector3f {
		return Vector3f(
			x.coerceAtLeast(minimal.x),
			y.coerceAtLeast(minimal.y),
			z.coerceAtLeast(minimal.z),
		)
	}

	override fun clamp(minimal: Vector3f, maximal: Vector3f): Vector3f {
		return Vector3f(
			x.coerceAtLeast(minimal.x).coerceAtMost(maximal.x),
			y.coerceAtLeast(minimal.y).coerceAtMost(maximal.y),
			z.coerceAtLeast(minimal.z).coerceAtMost(maximal.z),
		)
	}

	override fun unaryMinus(): Vector3f {
		return Vector3f(-x, -y, -z)
	}

	override fun distanceSquared(other: Vector3f): Double {
		val x = x.toDouble() - other.x.toDouble()
		val y = y.toDouble() - other.y.toDouble()
		val z = z.toDouble() - other.z.toDouble()

		return x * x + y * y + z * z
	}

	override val normalized: Vector3f
		get() {
			val len = length

			if (len == 0.0) {
				return ZERO
			}

			return Vector3f(
				(x / len).toFloat(),
				(y / len).toFloat(),
				(z / len).toFloat()
			)
		}

	override fun times(other: Float): Vector3f {
		return Vector3f(x * other, y * other, z * other)
	}

	override fun div(other: Float): Vector3f {
		return Vector3f(x / other, y / other, z / other)
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector4f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble() + other.z.toDouble() * z.toDouble()
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector3f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble() + other.z.toDouble() * z.toDouble()
	}

	/**
	 * Calculates scalar vector * vector, returning result as [Double]
	 */
	fun dot(other: Vector2f): Double {
		return other.x.toDouble() * x.toDouble() + other.y.toDouble() * y.toDouble()
	}

	/**
	 * Calculates vector vector * vector, returning result as new vector
	 */
	fun cross(other: Vector3f): Vector3f {
		return Vector3f(
			y * other.z - z * other.y,
			z * other.x - x * other.z,
			x * other.y - y * other.x,
		)
	}

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as Vector3f

		if (x != other.x) return false
		if (y != other.y) return false
		if (z != other.z) return false

		return true
	}

	override fun hashCode(): Int {
		var result = x.hashCode()
		result = 31 * result + y.hashCode()
		result = 31 * result + z.hashCode()
		return result
	}

	override fun toString(): String {
		return "[${x}f ${y}f ${z}f]"
	}

	inline val xx get() = Vector2f(x, x)
	inline val xy get() = Vector2f(x, y)
	inline val xz get() = Vector2f(x, z)
	inline val yx get() = Vector2f(y, x)
	inline val yy get() = Vector2f(y, y)
	inline val yz get() = Vector2f(y, z)
	inline val zx get() = Vector2f(z, x)
	inline val zy get() = Vector2f(z, y)
	inline val zz get() = Vector2f(z, z)
	inline val xxx get() = Vector3f(x, x, x)
	inline val xxy get() = Vector3f(x, x, y)
	inline val xxz get() = Vector3f(x, x, z)
	inline val xyx get() = Vector3f(x, y, x)
	inline val xyy get() = Vector3f(x, y, y)
	inline val xyz get() = Vector3f(x, y, z)
	inline val xzx get() = Vector3f(x, z, x)
	inline val xzy get() = Vector3f(x, z, y)
	inline val xzz get() = Vector3f(x, z, z)
	inline val yxx get() = Vector3f(y, x, x)
	inline val yxy get() = Vector3f(y, x, y)
	inline val yxz get() = Vector3f(y, x, z)
	inline val yyx get() = Vector3f(y, y, x)
	inline val yyy get() = Vector3f(y, y, y)
	inline val yyz get() = Vector3f(y, y, z)
	inline val yzx get() = Vector3f(y, z, x)
	inline val yzy get() = Vector3f(y, z, y)
	inline val yzz get() = Vector3f(y, z, z)
	inline val zxx get() = Vector3f(z, x, x)
	inline val zxy get() = Vector3f(z, x, y)
	inline val zxz get() = Vector3f(z, x, z)
	inline val zyx get() = Vector3f(z, y, x)
	inline val zyy get() = Vector3f(z, y, y)
	inline val zyz get() = Vector3f(z, y, z)
	inline val zzx get() = Vector3f(z, z, x)
	inline val zzy get() = Vector3f(z, z, y)
	inline val zzz get() = Vector3f(z, z, z)
	inline val xxxx get() = Vector4f(x, x, x, x)
	inline val xxxy get() = Vector4f(x, x, x, y)
	inline val xxxz get() = Vector4f(x, x, x, z)
	inline val xxyx get() = Vector4f(x, x, y, x)
	inline val xxyy get() = Vector4f(x, x, y, y)
	inline val xxyz get() = Vector4f(x, x, y, z)
	inline val xxzx get() = Vector4f(x, x, z, x)
	inline val xxzy get() = Vector4f(x, x, z, y)
	inline val xxzz get() = Vector4f(x, x, z, z)
	inline val xyxx get() = Vector4f(x, y, x, x)
	inline val xyxy get() = Vector4f(x, y, x, y)
	inline val xyxz get() = Vector4f(x, y, x, z)
	inline val xyyx get() = Vector4f(x, y, y, x)
	inline val xyyy get() = Vector4f(x, y, y, y)
	inline val xyyz get() = Vector4f(x, y, y, z)
	inline val xyzx get() = Vector4f(x, y, z, x)
	inline val xyzy get() = Vector4f(x, y, z, y)
	inline val xyzz get() = Vector4f(x, y, z, z)
	inline val xzxx get() = Vector4f(x, z, x, x)
	inline val xzxy get() = Vector4f(x, z, x, y)
	inline val xzxz get() = Vector4f(x, z, x, z)
	inline val xzyx get() = Vector4f(x, z, y, x)
	inline val xzyy get() = Vector4f(x, z, y, y)
	inline val xzyz get() = Vector4f(x, z, y, z)
	inline val xzzx get() = Vector4f(x, z, z, x)
	inline val xzzy get() = Vector4f(x, z, z, y)
	inline val xzzz get() = Vector4f(x, z, z, z)
	inline val yxxx get() = Vector4f(y, x, x, x)
	inline val yxxy get() = Vector4f(y, x, x, y)
	inline val yxxz get() = Vector4f(y, x, x, z)
	inline val yxyx get() = Vector4f(y, x, y, x)
	inline val yxyy get() = Vector4f(y, x, y, y)
	inline val yxyz get() = Vector4f(y, x, y, z)
	inline val yxzx get() = Vector4f(y, x, z, x)
	inline val yxzy get() = Vector4f(y, x, z, y)
	inline val yxzz get() = Vector4f(y, x, z, z)
	inline val yyxx get() = Vector4f(y, y, x, x)
	inline val yyxy get() = Vector4f(y, y, x, y)
	inline val yyxz get() = Vector4f(y, y, x, z)
	inline val yyyx get() = Vector4f(y, y, y, x)
	inline val yyyy get() = Vector4f(y, y, y, y)
	inline val yyyz get() = Vector4f(y, y, y, z)
	inline val yyzx get() = Vector4f(y, y, z, x)
	inline val yyzy get() = Vector4f(y, y, z, y)
	inline val yyzz get() = Vector4f(y, y, z, z)
	inline val yzxx get() = Vector4f(y, z, x, x)
	inline val yzxy get() = Vector4f(y, z, x, y)
	inline val yzxz get() = Vector4f(y, z, x, z)
	inline val yzyx get() = Vector4f(y, z, y, x)
	inline val yzyy get() = Vector4f(y, z, y, y)
	inline val yzyz get() = Vector4f(y, z, y, z)
	inline val yzzx get() = Vector4f(y, z, z, x)
	inline val yzzy get() = Vector4f(y, z, z, y)
	inline val yzzz get() = Vector4f(y, z, z, z)
	inline val zxxx get() = Vector4f(z, x, x, x)
	inline val zxxy get() = Vector4f(z, x, x, y)
	inline val zxxz get() = Vector4f(z, x, x, z)
	inline val zxyx get() = Vector4f(z, x, y, x)
	inline val zxyy get() = Vector4f(z, x, y, y)
	inline val zxyz get() = Vector4f(z, x, y, z)
	inline val zxzx get() = Vector4f(z, x, z, x)
	inline val zxzy get() = Vector4f(z, x, z, y)
	inline val zxzz get() = Vector4f(z, x, z, z)
	inline val zyxx get() = Vector4f(z, y, x, x)
	inline val zyxy get() = Vector4f(z, y, x, y)
	inline val zyxz get() = Vector4f(z, y, x, z)
	inline val zyyx get() = Vector4f(z, y, y, x)
	inline val zyyy get() = Vector4f(z, y, y, y)
	inline val zyyz get() = Vector4f(z, y, y, z)
	inline val zyzx get() = Vector4f(z, y, z, x)
	inline val zyzy get() = Vector4f(z, y, z, y)
	inline val zyzz get() = Vector4f(z, y, z, z)
	inline val zzxx get() = Vector4f(z, z, x, x)
	inline val zzxy get() = Vector4f(z, z, x, y)
	inline val zzxz get() = Vector4f(z, z, x, z)
	inline val zzyx get() = Vector4f(z, z, y, x)
	inline val zzyy get() = Vector4f(z, z, y, y)
	inline val zzyz get() = Vector4f(z, z, y, z)
	inline val zzzx get() = Vector4f(z, z, z, x)
	inline val zzzy get() = Vector4f(z, z, z, y)
	inline val zzzz get() = Vector4f(z, z, z, z)
	inline val rr get() = Vector2f(r, r)
	inline val rg get() = Vector2f(r, g)
	inline val rb get() = Vector2f(r, b)
	inline val gr get() = Vector2f(g, r)
	inline val gg get() = Vector2f(g, g)
	inline val gb get() = Vector2f(g, b)
	inline val br get() = Vector2f(b, r)
	inline val bg get() = Vector2f(b, g)
	inline val bb get() = Vector2f(b, b)
	inline val rrr get() = Vector3f(r, r, r)
	inline val rrg get() = Vector3f(r, r, g)
	inline val rrb get() = Vector3f(r, r, b)
	inline val rgr get() = Vector3f(r, g, r)
	inline val rgg get() = Vector3f(r, g, g)
	inline val rgb get() = Vector3f(r, g, b)
	inline val rbr get() = Vector3f(r, b, r)
	inline val rbg get() = Vector3f(r, b, g)
	inline val rbb get() = Vector3f(r, b, b)
	inline val grr get() = Vector3f(g, r, r)
	inline val grg get() = Vector3f(g, r, g)
	inline val grb get() = Vector3f(g, r, b)
	inline val ggr get() = Vector3f(g, g, r)
	inline val ggg get() = Vector3f(g, g, g)
	inline val ggb get() = Vector3f(g, g, b)
	inline val gbr get() = Vector3f(g, b, r)
	inline val gbg get() = Vector3f(g, b, g)
	inline val gbb get() = Vector3f(g, b, b)
	inline val brr get() = Vector3f(b, r, r)
	inline val brg get() = Vector3f(b, r, g)
	inline val brb get() = Vector3f(b, r, b)
	inline val bgr get() = Vector3f(b, g, r)
	inline val bgg get() = Vector3f(b, g, g)
	inline val bgb get() = Vector3f(b, g, b)
	inline val bbr get() = Vector3f(b, b, r)
	inline val bbg get() = Vector3f(b, b, g)
	inline val bbb get() = Vector3f(b, b, b)
	inline val rrrr get() = Vector4f(r, r, r, r)
	inline val rrrg get() = Vector4f(r, r, r, g)
	inline val rrrb get() = Vector4f(r, r, r, b)
	inline val rrgr get() = Vector4f(r, r, g, r)
	inline val rrgg get() = Vector4f(r, r, g, g)
	inline val rrgb get() = Vector4f(r, r, g, b)
	inline val rrbr get() = Vector4f(r, r, b, r)
	inline val rrbg get() = Vector4f(r, r, b, g)
	inline val rrbb get() = Vector4f(r, r, b, b)
	inline val rgrr get() = Vector4f(r, g, r, r)
	inline val rgrg get() = Vector4f(r, g, r, g)
	inline val rgrb get() = Vector4f(r, g, r, b)
	inline val rggr get() = Vector4f(r, g, g, r)
	inline val rggg get() = Vector4f(r, g, g, g)
	inline val rggb get() = Vector4f(r, g, g, b)
	inline val rgbr get() = Vector4f(r, g, b, r)
	inline val rgbg get() = Vector4f(r, g, b, g)
	inline val rgbb get() = Vector4f(r, g, b, b)
	inline val rbrr get() = Vector4f(r, b, r, r)
	inline val rbrg get() = Vector4f(r, b, r, g)
	inline val rbrb get() = Vector4f(r, b, r, b)
	inline val rbgr get() = Vector4f(r, b, g, r)
	inline val rbgg get() = Vector4f(r, b, g, g)
	inline val rbgb get() = Vector4f(r, b, g, b)
	inline val rbbr get() = Vector4f(r, b, b, r)
	inline val rbbg get() = Vector4f(r, b, b, g)
	inline val rbbb get() = Vector4f(r, b, b, b)
	inline val grrr get() = Vector4f(g, r, r, r)
	inline val grrg get() = Vector4f(g, r, r, g)
	inline val grrb get() = Vector4f(g, r, r, b)
	inline val grgr get() = Vector4f(g, r, g, r)
	inline val grgg get() = Vector4f(g, r, g, g)
	inline val grgb get() = Vector4f(g, r, g, b)
	inline val grbr get() = Vector4f(g, r, b, r)
	inline val grbg get() = Vector4f(g, r, b, g)
	inline val grbb get() = Vector4f(g, r, b, b)
	inline val ggrr get() = Vector4f(g, g, r, r)
	inline val ggrg get() = Vector4f(g, g, r, g)
	inline val ggrb get() = Vector4f(g, g, r, b)
	inline val gggr get() = Vector4f(g, g, g, r)
	inline val gggg get() = Vector4f(g, g, g, g)
	inline val gggb get() = Vector4f(g, g, g, b)
	inline val ggbr get() = Vector4f(g, g, b, r)
	inline val ggbg get() = Vector4f(g, g, b, g)
	inline val ggbb get() = Vector4f(g, g, b, b)
	inline val gbrr get() = Vector4f(g, b, r, r)
	inline val gbrg get() = Vector4f(g, b, r, g)
	inline val gbrb get() = Vector4f(g, b, r, b)
	inline val gbgr get() = Vector4f(g, b, g, r)
	inline val gbgg get() = Vector4f(g, b, g, g)
	inline val gbgb get() = Vector4f(g, b, g, b)
	inline val gbbr get() = Vector4f(g, b, b, r)
	inline val gbbg get() = Vector4f(g, b, b, g)
	inline val gbbb get() = Vector4f(g, b, b, b)
	inline val brrr get() = Vector4f(b, r, r, r)
	inline val brrg get() = Vector4f(b, r, r, g)
	inline val brrb get() = Vector4f(b, r, r, b)
	inline val brgr get() = Vector4f(b, r, g, r)
	inline val brgg get() = Vector4f(b, r, g, g)
	inline val brgb get() = Vector4f(b, r, g, b)
	inline val brbr get() = Vector4f(b, r, b, r)
	inline val brbg get() = Vector4f(b, r, b, g)
	inline val brbb get() = Vector4f(b, r, b, b)
	inline val bgrr get() = Vector4f(b, g, r, r)
	inline val bgrg get() = Vector4f(b, g, r, g)
	inline val bgrb get() = Vector4f(b, g, r, b)
	inline val bggr get() = Vector4f(b, g, g, r)
	inline val bggg get() = Vector4f(b, g, g, g)
	inline val bggb get() = Vector4f(b, g, g, b)
	inline val bgbr get() = Vector4f(b, g, b, r)
	inline val bgbg get() = Vector4f(b, g, b, g)
	inline val bgbb get() = Vector4f(b, g, b, b)
	inline val bbrr get() = Vector4f(b, b, r, r)
	inline val bbrg get() = Vector4f(b, b, r, g)
	inline val bbrb get() = Vector4f(b, b, r, b)
	inline val bbgr get() = Vector4f(b, b, g, r)
	inline val bbgg get() = Vector4f(b, b, g, g)
	inline val bbgb get() = Vector4f(b, b, g, b)
	inline val bbbr get() = Vector4f(b, b, b, r)
	inline val bbbg get() = Vector4f(b, b, b, g)
	inline val bbbb get() = Vector4f(b, b, b, b)
	inline val ss get() = Vector2f(s, s)
	inline val st get() = Vector2f(s, t)
	inline val sp get() = Vector2f(s, p)
	inline val ts get() = Vector2f(t, s)
	inline val tt get() = Vector2f(t, t)
	inline val tp get() = Vector2f(t, p)
	inline val ps get() = Vector2f(p, s)
	inline val pt get() = Vector2f(p, t)
	inline val pp get() = Vector2f(p, p)
	inline val sss get() = Vector3f(s, s, s)
	inline val sst get() = Vector3f(s, s, t)
	inline val ssp get() = Vector3f(s, s, p)
	inline val sts get() = Vector3f(s, t, s)
	inline val stt get() = Vector3f(s, t, t)
	inline val stp get() = Vector3f(s, t, p)
	inline val sps get() = Vector3f(s, p, s)
	inline val spt get() = Vector3f(s, p, t)
	inline val spp get() = Vector3f(s, p, p)
	inline val tss get() = Vector3f(t, s, s)
	inline val tst get() = Vector3f(t, s, t)
	inline val tsp get() = Vector3f(t, s, p)
	inline val tts get() = Vector3f(t, t, s)
	inline val ttt get() = Vector3f(t, t, t)
	inline val ttp get() = Vector3f(t, t, p)
	inline val tps get() = Vector3f(t, p, s)
	inline val tpt get() = Vector3f(t, p, t)
	inline val tpp get() = Vector3f(t, p, p)
	inline val pss get() = Vector3f(p, s, s)
	inline val pst get() = Vector3f(p, s, t)
	inline val psp get() = Vector3f(p, s, p)
	inline val pts get() = Vector3f(p, t, s)
	inline val ptt get() = Vector3f(p, t, t)
	inline val ptp get() = Vector3f(p, t, p)
	inline val pps get() = Vector3f(p, p, s)
	inline val ppt get() = Vector3f(p, p, t)
	inline val ppp get() = Vector3f(p, p, p)
	inline val ssss get() = Vector4f(s, s, s, s)
	inline val ssst get() = Vector4f(s, s, s, t)
	inline val sssp get() = Vector4f(s, s, s, p)
	inline val ssts get() = Vector4f(s, s, t, s)
	inline val sstt get() = Vector4f(s, s, t, t)
	inline val sstp get() = Vector4f(s, s, t, p)
	inline val ssps get() = Vector4f(s, s, p, s)
	inline val sspt get() = Vector4f(s, s, p, t)
	inline val sspp get() = Vector4f(s, s, p, p)
	inline val stss get() = Vector4f(s, t, s, s)
	inline val stst get() = Vector4f(s, t, s, t)
	inline val stsp get() = Vector4f(s, t, s, p)
	inline val stts get() = Vector4f(s, t, t, s)
	inline val sttt get() = Vector4f(s, t, t, t)
	inline val sttp get() = Vector4f(s, t, t, p)
	inline val stps get() = Vector4f(s, t, p, s)
	inline val stpt get() = Vector4f(s, t, p, t)
	inline val stpp get() = Vector4f(s, t, p, p)
	inline val spss get() = Vector4f(s, p, s, s)
	inline val spst get() = Vector4f(s, p, s, t)
	inline val spsp get() = Vector4f(s, p, s, p)
	inline val spts get() = Vector4f(s, p, t, s)
	inline val sptt get() = Vector4f(s, p, t, t)
	inline val sptp get() = Vector4f(s, p, t, p)
	inline val spps get() = Vector4f(s, p, p, s)
	inline val sppt get() = Vector4f(s, p, p, t)
	inline val sppp get() = Vector4f(s, p, p, p)
	inline val tsss get() = Vector4f(t, s, s, s)
	inline val tsst get() = Vector4f(t, s, s, t)
	inline val tssp get() = Vector4f(t, s, s, p)
	inline val tsts get() = Vector4f(t, s, t, s)
	inline val tstt get() = Vector4f(t, s, t, t)
	inline val tstp get() = Vector4f(t, s, t, p)
	inline val tsps get() = Vector4f(t, s, p, s)
	inline val tspt get() = Vector4f(t, s, p, t)
	inline val tspp get() = Vector4f(t, s, p, p)
	inline val ttss get() = Vector4f(t, t, s, s)
	inline val ttst get() = Vector4f(t, t, s, t)
	inline val ttsp get() = Vector4f(t, t, s, p)
	inline val ttts get() = Vector4f(t, t, t, s)
	inline val tttt get() = Vector4f(t, t, t, t)
	inline val tttp get() = Vector4f(t, t, t, p)
	inline val ttps get() = Vector4f(t, t, p, s)
	inline val ttpt get() = Vector4f(t, t, p, t)
	inline val ttpp get() = Vector4f(t, t, p, p)
	inline val tpss get() = Vector4f(t, p, s, s)
	inline val tpst get() = Vector4f(t, p, s, t)
	inline val tpsp get() = Vector4f(t, p, s, p)
	inline val tpts get() = Vector4f(t, p, t, s)
	inline val tptt get() = Vector4f(t, p, t, t)
	inline val tptp get() = Vector4f(t, p, t, p)
	inline val tpps get() = Vector4f(t, p, p, s)
	inline val tppt get() = Vector4f(t, p, p, t)
	inline val tppp get() = Vector4f(t, p, p, p)
	inline val psss get() = Vector4f(p, s, s, s)
	inline val psst get() = Vector4f(p, s, s, t)
	inline val pssp get() = Vector4f(p, s, s, p)
	inline val psts get() = Vector4f(p, s, t, s)
	inline val pstt get() = Vector4f(p, s, t, t)
	inline val pstp get() = Vector4f(p, s, t, p)
	inline val psps get() = Vector4f(p, s, p, s)
	inline val pspt get() = Vector4f(p, s, p, t)
	inline val pspp get() = Vector4f(p, s, p, p)
	inline val ptss get() = Vector4f(p, t, s, s)
	inline val ptst get() = Vector4f(p, t, s, t)
	inline val ptsp get() = Vector4f(p, t, s, p)
	inline val ptts get() = Vector4f(p, t, t, s)
	inline val pttt get() = Vector4f(p, t, t, t)
	inline val pttp get() = Vector4f(p, t, t, p)
	inline val ptps get() = Vector4f(p, t, p, s)
	inline val ptpt get() = Vector4f(p, t, p, t)
	inline val ptpp get() = Vector4f(p, t, p, p)
	inline val ppss get() = Vector4f(p, p, s, s)
	inline val ppst get() = Vector4f(p, p, s, t)
	inline val ppsp get() = Vector4f(p, p, s, p)
	inline val ppts get() = Vector4f(p, p, t, s)
	inline val pptt get() = Vector4f(p, p, t, t)
	inline val pptp get() = Vector4f(p, p, t, p)
	inline val ppps get() = Vector4f(p, p, p, s)
	inline val pppt get() = Vector4f(p, p, p, t)
	inline val pppp get() = Vector4f(p, p, p, p)

	companion object {
		@JvmField val ZERO = Vector3f()
		@JvmField val POSITIVE_X = Vector3f(x = 1f)
		@JvmField val NEGATIVE_X = Vector3f(x = -1f)
		@JvmField val POSITIVE_Y = Vector3f(y = 1f)
		@JvmField val NEGATIVE_Y = Vector3f(y = -1f)
		@JvmField val POSITIVE_Z = Vector3f(z = 1f)
		@JvmField val NEGATIVE_Z = Vector3f(z = -1f)
	}
}

operator fun Float.times(other: Vector3f): Vector3f {
	return Vector3f(this * other.x, this * other.y, this * other.z)
}

operator fun Float.div(other: Vector3f): Vector3f {
	return Vector3f(this / other.x, this / other.y, this / other.z)
}

/**
 * Mutable 3D Vector, representing three-dimensional coordinates as [Float]s
 *
 * It can be safely passed around to methods which expect immutable [Vector3f], **AND** do not store it,
 * as mutable operations are separated from immutable ones
 */
open class MutableVector3f(
	override var x: Float = 0f,
	override var y: Float = 0f,
	override var z: Float = 0f,
) : Vector3f(x, y), IMutableFloatVector<MutableVector3f>, IMutableFractionalVector<MutableVector3f>, IMutableVector<MutableVector3f, Vector3f>, IMatrixSetterFloat {
	constructor(input: IStruct2f, z: Float = 0f) : this(input.component1(), input.component2(), z)
	constructor(x: Float, input: IStruct2f) : this(x, input.component1(), input.component2())
	constructor(input: IStruct3f) : this(input.component1(), input.component2(), input.component3())

	override fun set(column: Int, row: Int, value: Float) {
		if (column != 0) {
			throw IndexOutOfBoundsException("Vectors are 1 column matrices ($column given)")
		}

		when (row) {
			0 -> x = value
			1 -> y = value
			2 -> z = value
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	override var r by this::x
	override var g by this::y
	override var b by this::z

	override var s by this::x
	override var t by this::y
	override var p by this::z

	override fun equals(other: Any?): Boolean {
		if (this === other) return true
		if (javaClass != other?.javaClass) return false

		other as MutableVector3f

		if (x != other.x) return false
		if (y != other.y) return false
		if (z != other.z) return false

		return true
	}

	override fun normalize(): Double {
		val len = length

		if (len == 0.0) {
			return 0.0
		}

		x = (x / len).toFloat()
		y = (y / len).toFloat()
		z = (z / len).toFloat()

		return len
	}

	override fun plusMut(other: Vector3f): MutableVector3f {
		x += other.x
		y += other.y
		z += other.z
		return this
	}

	override fun minusMut(other: Vector3f): MutableVector3f {
		x -= other.x
		y -= other.y
		z -= other.z
		return this
	}

	override fun timesMut(other: Vector3f): MutableVector3f {
		x *= other.x
		y *= other.y
		z *= other.z
		return this
	}

	override fun divMut(other: Vector3f): MutableVector3f {
		x /= other.x
		y /= other.y
		z /= other.z
		return this
	}

	override fun timesMut(other: Float): MutableVector3f {
		x *= other
		y *= other
		z *= other
		return this
	}

	override fun divMut(other: Float): MutableVector3f {
		x /= other
		y /= other
		z /= other
		return this
	}
}
