package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d

/**
 * The world class manages all physics entities, dynamic simulation,
 * and asynchronous queries. The world also contains efficient memory
 * management facilities.
 */
interface IB2World : IMovable {
	/**
	 * Register a destruction listener. The listener is owned by you and must
	 * remain in scope.
	 */
	var destructionListener: IDestructionListener?

	/**
	 * Register a contact filter to provide specific control over collision.
	 * Otherwise the default filter is used (b2_defaultFilter). The listener is
	 * owned by you and must remain in scope.
	 */
	var contactFilter: IContactFilter?

	/**
	 * Register a contact event listener. The listener is owned by you and must
	 * remain in scope.
	 */
	var contactListener: IContactListener?

	/**
	 * Register a routine for debug drawing. The debug draw functions are called
	 * inside with b2World::DebugDraw method. The debug draw object is owned
	 * by you and must remain in scope.
	 */
	var debugDraw: IDebugDraw?

	/**
	 * Create a rigid body given a definition. No reference to the definition
	 * is retained.
	 * @warning This function is locked during callbacks.
	 */
	fun createBody(bodyDef: ru.dbotthepony.kbox2d.api.BodyDef): ru.dbotthepony.kbox2d.api.IBody

	/**
	 * Destroy a rigid body given a definition. No reference to the definition
	 * is retained. This function is locked during callbacks.
	 * @warning This automatically deletes all associated shapes and joints.
	 * @warning This function is locked during callbacks.
	 */
	fun destroyBody(body: ru.dbotthepony.kbox2d.api.IBody)

	/**
	 * Create a joint to constrain bodies together. No reference to the definition
	 * is retained. This may cause the connected bodies to cease colliding.
	 * @warning This function is locked during callbacks.
	 */
	fun createJoint(jointDef: IJointDef): IJoint

	/**
	 * Destroy a joint. This may cause the connected bodies to begin colliding.
	 * @warning This function is locked during callbacks.
	 */
	fun destroyJoint(joint: IJoint)

	/**
	 * Take a time step. This performs collision detection, integration,
	 * and constraint solution.
	 * @param dt the amount of time to simulate, this should not vary.
	 * @param velocityIterations for the velocity constraint solver.
	 * @param positionIterations for the position constraint solver.
	 */
	fun step(dt: Double, velocityIterations: Int, positionIterations: Int)

	/**
	 * Manually clear the force buffer on all bodies. By default, forces are cleared automatically
	 * after each call to Step. The default behavior is modified by calling SetAutoClearForces.
	 * The purpose of this function is to support sub-stepping. Sub-stepping is often used to maintain
	 * a fixed sized time step under a variable frame-rate.
	 * When you perform sub-stepping you will disable auto clearing of forces and instead call
	 * ClearForces after all sub-steps are complete in one pass of your game loop.
	 * @see SetAutoClearForces
	 */
	fun clearForces()

	/**
	 * Call this to draw shapes and other debug draw data. This is intentionally non-const.
	 */
	fun debugDraw()

	/**
	 * Query the world for all fixtures that potentially overlap the
	 * provided AABB.
	 * @param callback a user implemented callback class.
	 * @param aabb the query box.
	 */
	fun queryAABB(aabb: AABB, callback: IQueryCallback)

	/**
	 * Ray-cast the world for all fixtures in the path of the ray. Your callback
	 * controls whether you get the closest point, any point, or n-points.
	 * The ray-cast ignores shapes that contain the starting point.
	 * @param callback a user implemented callback class.
	 * @param point1 the ray starting point
	 * @param point2 the ray ending point
	 */
	fun rayCast(point1: Vector2d, point2: Vector2d, callback: IRayCastCallback)

	/**
	 * Get the world body list. With the returned body, use b2Body::GetNext to get
	 * the next body in the world list. A nullptr body indicates the end of the list.
	 * @return the head of the world body list.
	 */
	val bodyList: ru.dbotthepony.kbox2d.api.IBody?

	val bodyListIterator: Iterator<ru.dbotthepony.kbox2d.api.IBody> get() {
		return object : Iterator<ru.dbotthepony.kbox2d.api.IBody> {
			private var node = bodyList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): ru.dbotthepony.kbox2d.api.IBody {
				val old = node!!
				node = old.next
				check(node != old) { "Hard loop detected at $old" }
				return old
			}
		}
	}

	/**
	 * Get the world joint list. With the returned joint, use b2Joint::GetNext to get
	 * the next joint in the world list. A nullptr joint indicates the end of the list.
	 * @return the head of the world joint list.
	 */
	val jointList: IJoint?

	val jointListIterator: Iterator<IJoint> get() {
		return object : Iterator<IJoint> {
			private var node = jointList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): IJoint {
				val old = node!!
				node = old.next
				check(node != old) { "Hard loop detected at $old" }
				return old
			}
		}
	}

	/**
	 * Get the world contact list. With the returned contact, use b2Contact::GetNext to get
	 * the next contact in the world list. A nullptr contact indicates the end of the list.
	 * @return the head of the world contact list.
	 * @warning contacts are created and destroyed in the middle of a time step.
	 * Use b2ContactListener to avoid missing contacts.
	 */
	val contactList: IContact? get() = contactManager.contactList

	val contactListIterator: Iterator<IContact> get() {
		return object : Iterator<IContact> {
			private var node = contactList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): IContact {
				val old = node!!
				node = old.next
				check(node != old) { "Hard loop detected at $old" }
				return old
			}
		}
	}

	/**
	 * Enable/disable sleep.
	 */
	var allowAutoSleep: Boolean

	/**
	 * Enable/disable warm starting. For testing.
	 */
	var warmStarting: Boolean

	/**
	 * Enable/disable continuous physics. For testing.
	 */
	var continuousPhysics: Boolean

	/**
	 * Enable/disable single stepped continuous physics. For testing.
	 */
	var enableSubStepping: Boolean

	val proxyCount: Int get() = contactManager.broadPhase.proxyCount
	val bodyCount: Int
	val jointCount: Int
	val contactCount: Int get() = contactManager.contactCount
	val treeHeight: Int get() = contactManager.broadPhase.treeHeight
	val treeBalance: Int get() = contactManager.broadPhase.treeBalance

	/**
	 * Get the quality metric of the dynamic tree. The smaller the better.
	 * The minimum is 1.
	 */
	val treeQuality: Double get() = contactManager.broadPhase.treeQuality

	/**
	 * Change the global gravity vector.
 	 */
	var gravity: Vector2d

	/**
	 * Is the world locked (in the middle of a time step).
	 */
	val isLocked: Boolean

	/**
	 * Set flag to control automatic clearing of forces after each time step.
	 */
	var autoClearForces: Boolean

	/**
	 * Get the current profile.
	 */
	val profileData: ru.dbotthepony.kbox2d.api.IProfileData

	/**
	 * Dump the world into the log file.
	 * @warning this should be called outside of a time step.
	 */
	fun dump()

	val contactManager: IContactManager

	fun notifyNewContacts()
}
