package ru.dbotthepony.kbox2d.collision

import it.unimi.dsi.fastutil.ints.IntArrayList
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d

class BroadPhase : IBroadPhase {
	private val moveBuffer = IntArrayList()
	private val pairBuffer = ArrayList<b2Pair>()
	private var moveCount = 0

	private val tree = DynamicTree()

	override var proxyCount: Int = 0
		private set

	override fun touchProxy(proxyID: Int) {
		bufferMove(proxyID)
	}

	override fun createProxy(aabb: AABB, userData: Any?): Int {
		val proxyId = tree.createProxy(aabb, userData)
		proxyCount++
		bufferMove(proxyId)
		return proxyId
	}

	override fun destroyProxy(proxyID: Int) {
		unBufferMove(proxyID)
		proxyCount--
		tree.destroyProxy(proxyID)
	}

	override fun moveProxy(proxyID: Int, aabb: AABB, displacement: Vector2d): Boolean {
		if (tree.moveProxy(proxyID, aabb, displacement)) {
			bufferMove(proxyID)
			return true
		}

		return false
	}

	private fun bufferMove(proxyId: Int) {
		if (moveBuffer.size > moveCount) {
			moveBuffer[moveCount] = proxyId
		} else {
			moveBuffer.add(proxyId)
		}

		moveCount++
	}

	private fun unBufferMove(proxyId: Int) {
		for (i in 0 until moveCount) {
			if (moveBuffer.getInt(i) == proxyId) {
				moveBuffer[i] = e_nullProxy
			}
		}
	}

	override fun query(aabb: AABB, callback: ProxyQueryCallback): Boolean {
		return tree.query(aabb, callback)
	}

	override fun rayCast(input: RayCastInput, callback: ProxyRayCastCallback) {
		tree.rayCast(input, callback)
	}

	override fun shiftOrigin(newOrigin: Vector2d) {
		tree.shiftOrigin(newOrigin)
	}

	override fun getUserData(proxyID: Int): Any? {
		return tree.getUserData(proxyID)
	}

	override fun getFatAABB(proxyID: Int): AABB {
		return tree.getFatAABB(proxyID)
	}

	override val treeHeight: Int
		get() = tree.height

	override val treeBalance: Int
		get() = tree.maxBalance

	override val treeQuality: Double
		get() = tree.getAreaRatio

	override fun testOverlap(proxyIDA: Int, proxyIDB: Int): Boolean {
		return tree.getFatAABB(proxyIDA).intersect(tree.getFatAABB(proxyIDB))
	}

	private var queryProxy: Int = -1

	private fun queryCallback(proxyId: Int, userData: Any?): Boolean {
		if (proxyId == queryProxy) {
			return true
		}

		val moved = tree.wasMoved(proxyId)

		if (moved && proxyId > queryProxy) {
			// Both proxies are moving. Avoid duplicate pairs.
			return true
		}

		if (proxyId > queryProxy) {
			pairBuffer.add(queryProxy to proxyId)
		} else {
			pairBuffer.add(proxyId to queryProxy)
		}

		return true
	}

	override fun updatePairs(callback: (Any?, Any?) -> Unit) {
		pairBuffer.clear()

		for (i in 0 until moveCount) {
			val value = moveBuffer.getInt(i)

			if (value == e_nullProxy)
				continue

			val fatAABB = tree.getFatAABB(value)
			queryProxy = value
			tree.query(fatAABB, this::queryCallback)
		}

		// Send pairs to caller
		for (primaryPair in pairBuffer) {
			val userDataA = tree.getUserData(primaryPair.first)
			val userDataB = tree.getUserData(primaryPair.second)

			callback.invoke(userDataA, userDataB)
		}

		// Clear move flags
		for (i in 0 until moveCount) {
			val value = moveBuffer.getInt(i)

			if (value != e_nullProxy) {
				tree.clearMoved(value)
			}
		}

		// Reset move buffer
		moveCount = 0
	}
}
