package ru.dbotthepony.kstarbound.math

import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.api.IStruct2d
import ru.dbotthepony.kstarbound.world.ChunkPos
import kotlin.math.absoluteValue

data class IntersectionTime(
	val invEntry: Vector2d,
	val invExit: Vector2d,
	val entry: Vector2d,
	val exit: Vector2d,
) {
	companion object {
		val ZERO = IntersectionTime(Vector2d.ZERO, Vector2d.ZERO, Vector2d.ZERO, Vector2d.ZERO)
	}
}

data class SweepResult(
	val normal: Vector2d,
	val collisionTime: Double,
	val intersectionTime: IntersectionTime
) {
	companion object {
		val ZERO = SweepResult(Vector2d.ZERO, 1.0, IntersectionTime.ZERO)
		val INTERSECT = SweepResult(Vector2d.ZERO, 0.0, IntersectionTime.ZERO)
	}
}

/**
 * Класс для описания Axis Aligned Bounding Box, двумя векторами,
 * где [mins] - нижняя левая точка,
 * [maxs] - верхняя правая
 */
data class AABB(val mins: Vector2d, val maxs: Vector2d) {
	init {
		require(mins.x <= maxs.x) { "mins.x ${mins.x} is more than maxs.x ${maxs.x}" }
		require(mins.y <= maxs.y) { "mins.y ${mins.y} is more than maxs.y ${maxs.y}" }
	}

	operator fun plus(other: AABB) = AABB(mins + other.mins, maxs + other.maxs)
	operator fun minus(other: AABB) = AABB(mins - other.mins, maxs - other.maxs)
	operator fun times(other: AABB) = AABB(mins * other.mins, maxs * other.maxs)
	operator fun div(other: AABB) = AABB(mins / other.mins, maxs / other.maxs)

	operator fun plus(other: Vector2d) = AABB(mins + other, maxs + other)
	operator fun minus(other: Vector2d) = AABB(mins - other, maxs - other)
	operator fun times(other: Vector2d) = AABB(mins * other, maxs * other)
	operator fun div(other: Vector2d) = AABB(mins / other, maxs / other)

	operator fun plus(other: Double) = AABB(mins + other, maxs + other)
	operator fun minus(other: Double) = AABB(mins - other, maxs - other)
	operator fun times(other: Double) = AABB(mins * other, maxs * other)
	operator fun div(other: Double) = AABB(mins / other, maxs / other)

	val xSpan get() = maxs.x - mins.x
	val ySpan get() = maxs.y - mins.y
	val centre get() = (mins + maxs) * 0.5

	val A get() = mins
	val B get() = Vector2d(mins.x, maxs.y)
	val C get() = maxs
	val D get() = Vector2d(maxs.x, mins.y)

	val bottomLeft get() = A
	val topLeft get() = B
	val topRight get() = C
	val bottomRight get() = D

	val width get() = maxs.x - mins.x
	val height get() = maxs.y - mins.y

	val extents get() = Vector2d(width * 0.5, height * 0.5)

	val diameter get() = mins.distance(maxs)
	val radius get() = diameter / 2.0

	val perimeter get() = (xSpan + ySpan) * 2.0

	fun isInside(point: Vector2d): Boolean {
		return point.x in mins.x .. maxs.x && point.y in mins.y .. maxs.y
	}

	/**
	 * Есть ли пересечение между этим AABB и [other]
	 *
	 * Считается, что они пересекаются, даже если у них просто равна одна из осей
	 */
	fun intersect(other: AABB): Boolean {
		val intersectX: Boolean

		if (xSpan <= other.xSpan)
			intersectX = mins.x in other.mins.x .. other.maxs.x || maxs.x in other.mins.x .. other.maxs.x
		else
			intersectX = other.mins.x in mins.x .. maxs.x || other.maxs.x in mins.x .. maxs.x

		if (!intersectX)
			return false

		val intersectY: Boolean

		if (ySpan <= other.ySpan)
			intersectY = mins.y in other.mins.y .. other.maxs.y || maxs.y in other.mins.y .. other.maxs.y
		else
			intersectY = other.mins.y in mins.y .. maxs.y || other.maxs.y in mins.y .. maxs.y

		return intersectY
	}

	/**
	 * @return Находится ли [other] внутри этого AABB
	 */
	fun contains(other: AABB): Boolean {
		if (xSpan < other.xSpan || ySpan < other.ySpan)
			return false

		return  other.mins.x in mins.x .. maxs.x &&
				other.maxs.x in mins.x .. maxs.x &&
				other.mins.y in mins.y .. maxs.y &&
				other.maxs.y in mins.y .. maxs.y
	}

	/**
	 * Есть ли пересечение между этим AABB и [other]
	 *
	 * Считается, что они НЕ пересекаются, если у них просто равна одна из осей
	 */
	fun intersectWeak(other: AABB): Boolean {
		if (maxs.x == other.mins.x || mins.x == other.maxs.x || maxs.y == other.mins.y || mins.y == other.maxs.y)
			return false

		val intersectX: Boolean

		if (xSpan <= other.xSpan)
			intersectX = mins.x in other.mins.x .. other.maxs.x || maxs.x in other.mins.x .. other.maxs.x
		else
			intersectX = other.mins.x in mins.x .. maxs.x || other.maxs.x in mins.x .. maxs.x

		if (!intersectX)
			return false

		val intersectY: Boolean

		if (ySpan <= other.ySpan)
			intersectY = mins.y in other.mins.y .. other.maxs.y || maxs.y in other.mins.y .. other.maxs.y
		else
			intersectY = other.mins.y in mins.y .. maxs.y || other.maxs.y in mins.y .. maxs.y

		return intersectY
	}

	fun intersectionDepth(other: AABB): Vector2d {
		val xDepth: Double
		val yDepth: Double

		val thisCentre = centre
		val otherCentre = other.centre

		if (thisCentre.x > otherCentre.x) {
			// считаем, что мы вошли справа
			xDepth = mins.x - other.maxs.x
		} else {
			// считаем, что мы вошли слева
			xDepth = maxs.x - other.mins.x
		}

		if (thisCentre.y > otherCentre.y) {
			// считаем, что мы вошли сверху
			yDepth = mins.y - other.maxs.y
		} else {
			// считаем, что мы вошли снизу
			yDepth = maxs.x - other.mins.x
		}

		return Vector2d(xDepth, yDepth)
	}

	fun distance(other: AABB): Double {
		val intersectX: Boolean
		val intersectY: Boolean

		if (ySpan <= other.ySpan)
			intersectY = mins.y in other.mins.y .. other.maxs.y || maxs.y in other.mins.y .. other.maxs.y
		else
			intersectY = other.mins.y in mins.y .. maxs.y || other.maxs.y in mins.y .. maxs.y

		if (xSpan <= other.xSpan)
			intersectX = mins.x in other.mins.x .. other.maxs.x || maxs.x in other.mins.x .. other.maxs.x
		else
			intersectX = other.mins.x in mins.x .. maxs.x || other.maxs.x in mins.x .. maxs.x

		if (intersectY && intersectX) {
			return 0.0
		}

		if (intersectX) {
			return (mins.y - other.maxs.y).absoluteValue.coerceAtMost((maxs.y - other.mins.y).absoluteValue)
		} else {
			return (mins.x - other.maxs.x).absoluteValue.coerceAtMost((maxs.x - other.mins.x).absoluteValue)
		}
	}

	fun pushOutFrom(other: AABB): Vector2d {
		if (!intersect(other))
			return Vector2d.ZERO

		val depth = intersectionDepth(other)

		if (depth.x.absoluteValue < depth.y.absoluteValue) {
			return Vector2d(x = depth.x)
		} else {
			return Vector2d(y = depth.y)
		}
	}

	/**
	 * Рассчитывает "время" пересечения
	 *
	 * https://www.gamedev.net/tutorials/programming/general-and-gameplay-programming/swept-aabb-collision-detection-and-response-r3084/
	 *
	 * Исправленный комментатором той же статьи от hypernewbie
	 */
	fun intersectionTime(other: AABB, velocity: Vector2d): IntersectionTime {
		val xInvEntry: Double
		val yInvEntry: Double
		val xInvExit: Double
		val yInvExit: Double

		if (velocity.x > 0.0) {
			xInvEntry = other.mins.x - maxs.x
			xInvExit = other.maxs.x - mins.x
		} else {
			xInvEntry = other.maxs.x - mins.x
			xInvExit = other.mins.x - maxs.x
		}

		if (velocity.y > 0.0) {
			yInvEntry = other.mins.y - maxs.y
			yInvExit = other.maxs.y - mins.y
		} else {
			yInvEntry = other.maxs.y - mins.y
			yInvExit = other.mins.y - maxs.y
		}

		var xEntry: Double
		var yEntry: Double
		val xExit: Double
		val yExit: Double

		if (velocity.x == 0.0) {
			xEntry = Double.NEGATIVE_INFINITY
			xExit = Double.POSITIVE_INFINITY
		} else {
			xEntry = xInvEntry / velocity.x
			xExit = xInvExit / velocity.x
		}

		if (velocity.y == 0.0) {
			yEntry = Double.NEGATIVE_INFINITY
			yExit = Double.POSITIVE_INFINITY
		} else {
			yEntry = yInvEntry / velocity.y
			yExit = yInvExit / velocity.y
		}

		if (yEntry > 1.0) yEntry = Double.NEGATIVE_INFINITY
		if (xEntry > 1.0) xEntry = Double.NEGATIVE_INFINITY

		return IntersectionTime(
			Vector2d(xInvEntry, yInvEntry),
			Vector2d(xInvExit, yInvExit),
			Vector2d(xEntry, yEntry),
			Vector2d(xExit, yExit),
		)
	}

	/**
	 * Рассчитывает нормаль пересечения и процент пути ("время"), на котором произошло столкновение.
	 *
	 * Если столкновение не произошло, то возвращается [SweepResult.ZERO]
	 *
	 * Внимание: Если пересечение уже произошло (т.е. другой AABB пересекается с this), то данный метод
	 * вернёт заведомо ложный результат (т.е. "нет пересечения")
	 *
	 * https://www.gamedev.net/tutorials/programming/general-and-gameplay-programming/swept-aabb-collision-detection-and-response-r3084/
	 *
	 * Исправленный комментатором той же статьи от hypernewbie
	 */
	fun sweep(other: AABB, velocity: Vector2d): SweepResult {
		val time = intersectionTime(other, velocity)
		val (near, far, entry, exit) = time
		val (xEntry, yEntry) = entry
		val (xExit, yExit) = exit

		val entryTime = xEntry.coerceAtLeast(yEntry)
		val exitTime = xExit.coerceAtLeast(yExit)

		// гарантированно нет столкновения
		if (entryTime > exitTime || xEntry < 0.0 && yEntry < 0.0) {
			return SweepResult.ZERO
		}

		if (xEntry < 0.0) {
			if (maxs.x < other.mins.x || mins.x > other.maxs.x)
				return SweepResult.ZERO
		}

		if (yEntry < 0.0) {
			if (maxs.y < other.mins.y || mins.y > other.maxs.y)
				return SweepResult.ZERO
		}

		val (xInvEntry, yInvEntry) = near
		val normal: Vector2d

		if (xEntry > yEntry) {
			if (xInvEntry <= 0.0) { // исправление от меня: <= для тех случаев, когда мы уже на той же оси
				normal = Vector2d.RIGHT
			} else {
				normal = Vector2d.LEFT
			}
		} else {
			if (yInvEntry <= 0.0) { // исправление от меня: <= для тех случаев, когда мы уже на той же оси
				normal = Vector2d.UP
			} else {
				normal = Vector2d.DOWN
			}
		}

		return SweepResult(normal, entryTime, time)
	}

	/**
	 * Рассчитывает нормаль пересечения и процент пути ("время"), на котором произошло столкновение.
	 *
	 * Если столкновение не произошло, то возвращается [SweepResult.ZERO]
	 *
	 * Если данный AABB уже столкнулся с [other], возвращается [SweepResult.INTERSECT]
	 *
	 * https://www.gamedev.net/tutorials/programming/general-and-gameplay-programming/swept-aabb-collision-detection-and-response-r3084/
	 */
	fun safeSweep(other: AABB, velocity: Vector2d): SweepResult {
		if (intersect(other)) {
			return SweepResult.INTERSECT
		}

		return sweep(other, velocity)
	}

	fun encasingIntAABB(): AABBi {
		return AABBi(
			Vector2i(roundByAbsoluteValue(mins.x), roundByAbsoluteValue(mins.y)),
			Vector2i(roundByAbsoluteValue(maxs.x), roundByAbsoluteValue(maxs.y)),
		)
	}

	fun encasingChunkPosAABB(): AABBi {
		return AABBi(
			Vector2i(ChunkPos.tileToChunkComponent(roundTowardsNegativeInfinity(mins.x)), ChunkPos.tileToChunkComponent(roundTowardsNegativeInfinity(mins.y))),
			Vector2i(ChunkPos.tileToChunkComponent(roundTowardsPositiveInfinity(maxs.x)), ChunkPos.tileToChunkComponent(roundTowardsPositiveInfinity(maxs.y))),
		)
	}

	/**
	 * Возвращает AABB, который содержит в себе оба AABB
	 */
	fun combine(other: AABB): AABB {
		val minX = mins.x.coerceAtMost(other.mins.x)
		val minY = mins.y.coerceAtMost(other.mins.y)
		val maxX = maxs.x.coerceAtLeast(other.maxs.x)
		val maxY = maxs.y.coerceAtLeast(other.maxs.y)

		return AABB(Vector2d(minX, minY), Vector2d(maxX, maxY))
	}

	companion object {
		fun rectangle(pos: IStruct2d, width: Double, height: Double = width): AABB {
			val (x, y) = pos

			return AABB(
				Vector2d(x - width / 2.0, y - height / 2.0),
				Vector2d(x + width / 2.0, y + height / 2.0),
			)
		}
	}
}

object AABBTypeAdapter : TypeAdapter<AABB>() {
	override fun write(out: JsonWriter, value: AABB) {
		`out`.beginArray()
		Vector2dTypeAdapter.write(out, value.mins)
		Vector2dTypeAdapter.write(out, value.maxs)
		`out`.endArray()
	}

	override fun read(`in`: JsonReader): AABB {
		val (x1, x2) = Vector2dTypeAdapter.read(`in`)
		val (y1, y2) = Vector2dTypeAdapter.read(`in`)

		val xMins = x1.coerceAtMost(x2)
		val xMaxs = x1.coerceAtLeast(x2)

		val yMins = y1.coerceAtMost(y2)
		val yMaxs = y1.coerceAtLeast(y2)

		return AABB(
			Vector2d(xMins, yMins),
			Vector2d(xMaxs, yMaxs),
		)
	}
}

data class AABBi(val mins: Vector2i, val maxs: Vector2i) {
	init {
		require(mins.x <= maxs.x) { "mins.x ${mins.x} is more than maxs.x ${maxs.x}" }
		require(mins.y <= maxs.y) { "mins.y ${mins.y} is more than maxs.y ${maxs.y}" }
	}

	operator fun plus(other: AABBi) = AABBi(mins + other.mins, maxs + other.maxs)
	operator fun minus(other: AABBi) = AABBi(mins - other.mins, maxs - other.maxs)
	operator fun times(other: AABBi) = AABBi(mins * other.mins, maxs * other.maxs)
	operator fun div(other: AABBi) = AABBi(mins / other.mins, maxs / other.maxs)

	operator fun plus(other: Vector2i) = AABBi(mins + other, maxs + other)
	operator fun minus(other: Vector2i) = AABBi(mins - other, maxs - other)
	operator fun times(other: Vector2i) = AABBi(mins * other, maxs * other)
	operator fun div(other: Vector2i) = AABBi(mins / other, maxs / other)

	operator fun plus(other: Int) = AABBi(mins + other, maxs + other)
	operator fun minus(other: Int) = AABBi(mins - other, maxs - other)
	operator fun times(other: Int) = AABBi(mins * other, maxs * other)
	operator fun div(other: Int) = AABBi(mins / other, maxs / other)

	val xSpan get() = maxs.x - mins.x
	val ySpan get() = maxs.y - mins.y
	val centre get() = mins.toDoubleVector() + maxs.toDoubleVector() * 0.5

	val A get() = mins
	val B get() = Vector2i(mins.x, maxs.y)
	val C get() = maxs
	val D get() = Vector2i(maxs.x, mins.y)

	val bottomLeft get() = A
	val topLeft get() = B
	val topRight get() = C
	val bottomRight get() = D

	val width get() = (maxs.x - mins.x) / 2
	val height get() = (maxs.y - mins.y) / 2

	val diameter get() = mins.distance(maxs)
	val radius get() = diameter / 2.0

	val perimeter get() = (xSpan + ySpan) * 2

	fun isInside(point: Vector2i): Boolean {
		return point.x in mins.x .. maxs.x && point.y in mins.y .. maxs.y
	}

	/**
	 * Есть ли пересечение между этим AABB и [other]
	 *
	 * Считается, что они пересекаются, даже если у них просто равна одна из осей
	 */
	fun intersect(other: AABBi): Boolean {
		val intersectX: Boolean

		if (xSpan <= other.xSpan)
			intersectX = mins.x in other.mins.x .. other.maxs.x || maxs.x in other.mins.x .. other.maxs.x
		else
			intersectX = other.mins.x in mins.x .. maxs.x || other.maxs.x in mins.x .. maxs.x

		if (!intersectX)
			return false

		val intersectY: Boolean

		if (ySpan <= other.ySpan)
			intersectY = mins.y in other.mins.y .. other.maxs.y || maxs.y in other.mins.y .. other.maxs.y
		else
			intersectY = other.mins.y in mins.y .. maxs.y || other.maxs.y in mins.y .. maxs.y

		return intersectY
	}

	/**
	 * Есть ли пересечение между этим AABB и [other]
	 *
	 * Считается, что они НЕ пересекаются, если у них просто равна одна из осей
	 */
	fun intersectWeak(other: AABBi): Boolean {
		if (maxs.x == other.mins.x || mins.x == other.maxs.x || maxs.y == other.mins.y || mins.y == other.maxs.y)
			return false

		val intersectX: Boolean

		if (xSpan <= other.xSpan)
			intersectX = mins.x in other.mins.x .. other.maxs.x || maxs.x in other.mins.x .. other.maxs.x
		else
			intersectX = other.mins.x in mins.x .. maxs.x || other.maxs.x in mins.x .. maxs.x

		if (!intersectX)
			return false

		val intersectY: Boolean

		if (ySpan <= other.ySpan)
			intersectY = mins.y in other.mins.y .. other.maxs.y || maxs.y in other.mins.y .. other.maxs.y
		else
			intersectY = other.mins.y in mins.y .. maxs.y || other.maxs.y in mins.y .. maxs.y

		return intersectY
	}

	fun toDoubleAABB() = AABB(mins.toDoubleVector(), maxs.toDoubleVector())

	private inner class Iterator<T>(private val factory: (x: Int, y: Int) -> T) : kotlin.collections.Iterator<T> {
		private var x = mins.x
		private var y = mins.y
		private var next = true

		override fun hasNext(): Boolean {
			return next
		}

		override fun next(): T {
			if (!next)
				throw IllegalStateException()

			val obj = factory.invoke(x++, y)

			if (x > maxs.x) {
				x = mins.x

				if (++y > maxs.y) {
					next = false
				}
			}

			return obj
		}
	}

	val vectors: kotlin.collections.Iterator<Vector2i> get() = Iterator(::Vector2i)
	val chunkPositions: kotlin.collections.Iterator<ChunkPos> get() = Iterator(::ChunkPos)
}

object AABBiTypeAdapter : TypeAdapter<AABBi>() {
	override fun write(out: JsonWriter, value: AABBi) {
		`out`.beginArray()
		Vector2iTypeAdapter.write(out, value.mins)
		Vector2iTypeAdapter.write(out, value.maxs)
		`out`.endArray()
	}

	override fun read(`in`: JsonReader): AABBi {
		val (x1, x2) = Vector2iTypeAdapter.read(`in`)
		val (y1, y2) = Vector2iTypeAdapter.read(`in`)

		val xMins = x1.coerceAtMost(x2)
		val xMaxs = x1.coerceAtLeast(x2)

		val yMins = y1.coerceAtMost(y2)
		val yMaxs = y1.coerceAtLeast(y2)

		return AABBi(
			Vector2i(xMins, yMins),
			Vector2i(xMaxs, yMaxs),
		)
	}
}

