package ru.dbotthepony.kstarbound.math

import ru.dbotthepony.kbox2d.api.b2Cross
import ru.dbotthepony.kbox2d.api.b2Dot
import ru.dbotthepony.kstarbound.api.IStruct3d
import ru.dbotthepony.kstarbound.api.IStruct3f
import ru.dbotthepony.kstarbound.api.IStruct4d
import ru.dbotthepony.kstarbound.api.IStruct4f
import java.nio.ByteBuffer
import java.nio.ByteOrder
import java.nio.DoubleBuffer
import java.nio.FloatBuffer
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.tan

interface IMatrixLike {
	val columns: Int
	val rows: Int
}

interface IMatrixLikeInt {
	operator fun get(row: Int, column: Int): Int
}

interface IMatrixLikeFloat : IMatrixLike {
	operator fun get(row: Int, column: Int): Float
}

interface IMatrixLikeDouble : IMatrixLike {
	operator fun get(row: Int, column: Int): Double
}

interface IMatrix : IMatrixLike {
	operator fun plus(other: IMatrix): IMatrix
	operator fun minus(other: IMatrix): IMatrix
	operator fun times(other: IMatrix): IMatrix
}

interface FloatMatrix<T : FloatMatrix<T>> : IMatrix, IMatrixLikeFloat {
	operator fun plus(other: Float): T
	operator fun minus(other: Float): T
	operator fun times(other: Float): T
	operator fun div(other: Float): T

	override operator fun plus(other: IMatrix): T
	override operator fun minus(other: IMatrix): T
	override operator fun times(other: IMatrix): T

	/**
	 * Если матрица больше или меньше, считать что всевозможные остальные координаты равны единице (не менять)
	 */
	fun scale(x: Float, y: Float = 1f, z: Float = 1f, w: Float = 1f): T
	fun scale(value: IStruct4f): T {
		val (x, y, z, w) = value
		return scale(x, y, z, w)
	}

	fun translate(x: Float = 0f, y: Float = 0f, z: Float = 0f): T
	fun translate(value: IStruct3f): T {
		val (x, y, z) = value
		return translate(x, y, z)
	}

	fun translateWithScale(x: Float = 0f, y: Float = 0f, z: Float = 0f): T
	fun translateWithScale(value: IStruct3f): T {
		val (x, y, z) = value
		return translateWithScale(x, y, z)
	}

	/**
	 * Выдает массив в готовом для OpenGL виде (строка -> столбец) по умолчанию
	 */
	fun toFloatArray(columnMajor: Boolean = true): FloatArray {
		val buff = FloatArray(rows * columns)

		if (columnMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row + rows * column] = this[row, column]
				}
			}
		} else {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row * columns + column] = this[row, column]
				}
			}
		}

		return buff
	}
}

interface DoubleMatrix<T : DoubleMatrix<T>> : IMatrix, IMatrixLikeDouble {
	operator fun plus(other: Double): T
	operator fun minus(other: Double): T
	operator fun times(other: Double): T
	operator fun div(other: Double): T

	override operator fun plus(other: IMatrix): T
	override operator fun minus(other: IMatrix): T
	override operator fun times(other: IMatrix): T

	/**
	 * Если матрица больше или меньше, считать что всевозможные остальные координаты равны единице (не менять)
	 */
	fun scale(x: Double, y: Double = 1.0, z: Double = 1.0, w: Double = 1.0): T

	fun scale(value: IStruct4d): T {
		val (x, y, z, w) = value
		return scale(x, y, z, w)
	}

	fun translate(x: Double = 0.0, y: Double = 0.0, z: Double = 0.0): T

	fun translate(value: IStruct3d): T {
		val (x, y, z) = value
		return translate(x, y, z)
	}

	fun translateWithScale(x: Double = 0.0, y: Double = 0.0, z: Double = 0.0): T

	fun translateWithScale(value: IStruct3d): T {
		val (x, y, z) = value
		return translateWithScale(x, y, z)
	}

	/**
	 * Выдает массив в готовом для OpenGL виде (строка -> столбец) по умолчанию
	 */
	fun toDoubleArray(columnMajor: Boolean = true): DoubleArray {
		val buff = DoubleArray(rows * columns)

		if (columnMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row + rows * column] = this[row, column]
				}
			}
		} else {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					buff[row * columns + column] = this[row, column]
				}
			}
		}

		return buff
	}
}

interface MutableFloatMatrix<T : MutableFloatMatrix<T>> : FloatMatrix<T> {
	operator fun set(row: Int, column: Int, value: Float)
}

abstract class AbstractMatrix4f<T : AbstractMatrix4f<T>> : FloatMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 4 * 4).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectFloatBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toFloatBuffer(columnMajor: Boolean = true): FloatBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)
			directBuffer.put(m30)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)
			directBuffer.put(m31)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
			directBuffer.put(m32)

			directBuffer.put(m03)
			directBuffer.put(m13)
			directBuffer.put(m23)
			directBuffer.put(m33)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)
			directBuffer.put(m03)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)
			directBuffer.put(m13)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
			directBuffer.put(m23)

			directBuffer.put(m30)
			directBuffer.put(m31)
			directBuffer.put(m32)
			directBuffer.put(m33)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Float; abstract val m01: Float; abstract val m02: Float; abstract val m03: Float
	abstract val m10: Float; abstract val m11: Float; abstract val m12: Float; abstract val m13: Float
	abstract val m20: Float; abstract val m21: Float; abstract val m22: Float; abstract val m23: Float
	abstract val m30: Float; abstract val m31: Float; abstract val m32: Float; abstract val m33: Float

	override val columns: Int
		get() = 4

	override val rows: Int
		get() = 4

	override fun get(row: Int, column: Int): Float {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				3 -> m30
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				3 -> m31
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				3 -> m32
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03
				1 -> m13
				2 -> m23
				3 -> m33
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrixes")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is AbstractMatrix4f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02, this.m03 + m03,
			this.m10 + m10, this.m11 + m11, this.m12 + m12, this.m13 + m13,
			this.m20 + m20, this.m21 + m21, this.m22 + m22, this.m23 + m23,
			this.m30 + m30, this.m31 + m31, this.m32 + m32, this.m33 + m33,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is AbstractMatrix4f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02, this.m03 - m03,
			this.m10 - m10, this.m11 - m11, this.m12 - m12, this.m13 - m13,
			this.m20 - m20, this.m21 - m21, this.m22 - m22, this.m23 - m23,
			this.m30 - m30, this.m31 - m31, this.m32 - m32, this.m33 - m33,
		)
	}

	override fun plus(other: Float): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other, this.m03 + other,
			this.m10 + other, this.m11 + other, this.m12 + other, this.m13 + other,
			this.m20 + other, this.m21 + other, this.m22 + other, this.m23 + other,
			this.m30 + other, this.m31 + other, this.m32 + other, this.m33 + other,
		)
	}

	override fun minus(other: Float): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other, this.m03 - other,
			this.m10 - other, this.m11 - other, this.m12 - other, this.m13 - other,
			this.m20 - other, this.m21 - other, this.m22 - other, this.m23 - other,
			this.m30 - other, this.m31 - other, this.m32 - other, this.m33 - other,
		)
	}

	override fun times(other: Float): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other, this.m03 * other,
			this.m10 * other, this.m11 * other, this.m12 * other, this.m13 * other,
			this.m20 * other, this.m21 * other, this.m22 * other, this.m23 * other,
			this.m30 * other, this.m31 * other, this.m32 * other, this.m33 * other,
		)
	}

	override fun scale(x: Float, y: Float, z: Float, w: Float): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02, this.m03,
			this.m10, this.m11 * y, this.m12, this.m13,
			this.m20, this.m21, this.m22 * z, this.m23,
			this.m30, this.m31, this.m32, this.m33 * w,
		)
	}

	override fun translate(x: Float, y: Float, z: Float): T {
		return createOrModify(
			this.m00, this.m01, this.m02, this.m03 + x,
			this.m10, this.m11, this.m12, this.m13 + y,
			this.m20, this.m21, this.m22, this.m23 + z,
			this.m30, this.m31, this.m32, this.m33,
		)
	}

	open val translation: Vector3f
		get() {
			return Vector3f(m03, m13, m23)
		}

	override fun div(other: Float): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other, this.m03 / other,
			this.m10 / other, this.m11 / other, this.m12 / other, this.m13 / other,
			this.m20 / other, this.m21 / other, this.m22 / other, this.m23 / other,
			this.m30 / other, this.m31 / other, this.m32 / other, this.m33 / other,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4f can only use 4x4 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float; val m03: Float;
		val m10: Float; val m11: Float; val m12: Float; val m13: Float;
		val m20: Float; val m21: Float; val m22: Float; val m23: Float;
		val m30: Float; val m31: Float; val m32: Float; val m33: Float;

		if (other is AbstractMatrix4f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20 +
					this.m03 * m30

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20 +
					this.m13 * m30

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20 +
					this.m23 * m30

		val newm30 =
			this.m30 * m00 +
					this.m31 * m10 +
					this.m32 * m20 +
					this.m33 * m30

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21 +
					this.m03 * m31

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21 +
					this.m13 * m31

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21 +
					this.m23 * m31

		val newm31 =
			this.m30 * m01 +
					this.m31 * m11 +
					this.m32 * m21 +
					this.m33 * m31

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22 +
					this.m03 * m32

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22 +
					this.m13 * m32

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22 +
					this.m23 * m32

		val newm32 =
			this.m30 * m02 +
					this.m31 * m12 +
					this.m32 * m22 +
					this.m33 * m32

		// четвёртый столбец
		val newm03 =
			this.m00 * m03 +
					this.m01 * m13 +
					this.m02 * m23 +
					this.m03 * m33

		val newm13 =
			this.m10 * m03 +
					this.m11 * m13 +
					this.m12 * m23 +
					this.m13 * m33

		val newm23 =
			this.m20 * m03 +
					this.m21 * m13 +
					this.m22 * m23 +
					this.m23 * m33

		val newm33 =
			this.m30 * m03 +
					this.m31 * m13 +
					this.m32 * m23 +
					this.m33 * m33

		return createOrModify(
			newm00, newm01, newm02, newm03,
			newm10, newm11, newm12, newm13,
			newm20, newm21, newm22, newm23,
			newm30, newm31, newm32, newm33,
		)
	}

	override fun translateWithScale(x: Float, y: Float, z: Float): T {
		return createOrModify(
			m00, m01, m02, m03 + x * m00 + y * m01 + z * m02,
			m10, m11, m12, m13 + x * m10 + y * m11 + z * m12,
			m20, m21, m22, m23 + x * m20 + y * m21 + z * m22,
			m30, m31, m32, m33,
		)
	}

	val absoluteMatrix get() = Matrix4f(
		m00 = m00.absoluteValue, m01 = m01.absoluteValue, m02 = m02.absoluteValue, m03 = m03.absoluteValue,
		m10 = m10.absoluteValue, m11 = m11.absoluteValue, m12 = m12.absoluteValue, m13 = m13.absoluteValue,
		m20 = m20.absoluteValue, m21 = m21.absoluteValue, m22 = m22.absoluteValue, m23 = m23.absoluteValue,
		m30 = m30.absoluteValue, m31 = m31.absoluteValue, m32 = m32.absoluteValue, m33 = m33.absoluteValue,
	)
}

data class Matrix4f(
	override val m00: Float = 1f, override val m01: Float = 0f, override val m02: Float = 0f, override val m03: Float = 0f,
	override val m10: Float = 0f, override val m11: Float = 1f, override val m12: Float = 0f, override val m13: Float = 0f,
	override val m20: Float = 0f, override val m21: Float = 0f, override val m22: Float = 1f, override val m23: Float = 0f,
	override val m30: Float = 0f, override val m31: Float = 0f, override val m32: Float = 0f, override val m33: Float = 1f,
) : AbstractMatrix4f<Matrix4f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): Matrix4f {
		return Matrix4f(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	fun toMutableMatrix(): MutableMatrix4f {
		return MutableMatrix4f(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	companion object {
		val IDENTITY = Matrix4f()

		val SCREEN_FLIP = IDENTITY.let {
			return@let it * Vector3f.FORWARD.rotateAroundThis(-PI / 2)
		}

		/**
		 * Возвращает матрицу ортографической проекции, с ИНВЕНТИРОВАННОЙ y координатой, и с добавлением 2f
		 *
		 * Т.е. любое значение компоненты вектора y будет иметь противоположный знак после перемножения на данную матрицу
		 *
		 * Смысл данного изменения знака в преобразовании экранных координат OpenGL к вменяемому виду. Многие примеры указывают Z как отрицательную компоненту,
		 * что так же "убирает" это недоумение, только вот у нас Z это глубина, а не высота
		 *
		 * y = 0 будет соответствовать верхнему левому углу окна
		 */
		fun ortho(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return Matrix4f(
				m00 = 2f / (right - left),
				m11 = -2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom) + 2f,
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Возвращает матрицу ортографической проекции, без инвентирования
		 *
		 * y = 0 будет соответствовать нижнему левому углу окна
		 */
		fun orthoDirect(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return Matrix4f(
				m00 = 2f / (right - left),
				m11 = 2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom),
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		fun perspective(fov: Float, zFar: Float, zNear: Float): Matrix4f {
			val scale = (1.0 / (tan(Math.toRadians(fov.toDouble()) / 2.0))).toFloat()
			val r = zFar - zNear

			return Matrix4f(
				m00 = scale,
				m11 = scale,
				m22 = -zFar / r,
				m23 = -1f,
				m32 = -zFar * zNear / r,
			)
		}
	}
}

data class MutableMatrix4f(
	override var m00: Float = 1f, override var m01: Float = 0f, override var m02: Float = 0f, override var m03: Float = 0f,
	override var m10: Float = 0f, override var m11: Float = 1f, override var m12: Float = 0f, override var m13: Float = 0f,
	override var m20: Float = 0f, override var m21: Float = 0f, override var m22: Float = 1f, override var m23: Float = 0f,
	override var m30: Float = 0f, override var m31: Float = 0f, override var m32: Float = 0f, override var m33: Float = 1f,
) : AbstractMatrix4f<MutableMatrix4f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float, m03: Float,
		m10: Float, m11: Float, m12: Float, m13: Float,
		m20: Float, m21: Float, m22: Float, m23: Float,
		m30: Float, m31: Float, m32: Float, m33: Float,
	): MutableMatrix4f {
		this.m00 = m00; this.m01 = m01; this.m02 = m02; this.m03 = m03
		this.m10 = m10; this.m11 = m11; this.m12 = m12; this.m13 = m13
		this.m20 = m20; this.m21 = m21; this.m22 = m22; this.m23 = m23
		this.m30 = m30; this.m31 = m31; this.m32 = m32; this.m33 = m33

		return this
	}

	override var translation: Vector3f
		get() {
			return Vector3f(m03, m13, m23)
		}
		set(value) {
			translate(value)
		}

	fun set(row: Int, column: Int, value: Float) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				3 -> m30 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				3 -> m31 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				3 -> m32 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03 = value
				1 -> m13 = value
				2 -> m23 = value
				3 -> m33 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}

abstract class AbstractMatrix4d<T : AbstractMatrix4d<T>> : DoubleMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 4 * 4).also { it.order(ByteOrder.nativeOrder()) }.asDoubleBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectDoubleBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toDoubleBuffer(columnMajor: Boolean = true): DoubleBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)
			directBuffer.put(m30)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)
			directBuffer.put(m31)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
			directBuffer.put(m32)

			directBuffer.put(m03)
			directBuffer.put(m13)
			directBuffer.put(m23)
			directBuffer.put(m33)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)
			directBuffer.put(m03)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)
			directBuffer.put(m13)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
			directBuffer.put(m23)

			directBuffer.put(m30)
			directBuffer.put(m31)
			directBuffer.put(m32)
			directBuffer.put(m33)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Double; abstract val m01: Double; abstract val m02: Double; abstract val m03: Double
	abstract val m10: Double; abstract val m11: Double; abstract val m12: Double; abstract val m13: Double
	abstract val m20: Double; abstract val m21: Double; abstract val m22: Double; abstract val m23: Double
	abstract val m30: Double; abstract val m31: Double; abstract val m32: Double; abstract val m33: Double

	override val columns: Int
		get() = 4

	override val rows: Int
		get() = 4

	override fun get(row: Int, column: Int): Double {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				3 -> m30
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				3 -> m31
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				3 -> m32
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03
				1 -> m13
				2 -> m23
				3 -> m33
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Double, m01: Double, m02: Double, m03: Double,
		m10: Double, m11: Double, m12: Double, m13: Double,
		m20: Double, m21: Double, m22: Double, m23: Double,
		m30: Double, m31: Double, m32: Double, m33: Double,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4d can only use 4x4 matrixes")
		}

		val m00: Double; val m01: Double; val m02: Double; val m03: Double;
		val m10: Double; val m11: Double; val m12: Double; val m13: Double;
		val m20: Double; val m21: Double; val m22: Double; val m23: Double;
		val m30: Double; val m31: Double; val m32: Double; val m33: Double;

		if (other is AbstractMatrix4d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02, this.m03 + m03,
			this.m10 + m10, this.m11 + m11, this.m12 + m12, this.m13 + m13,
			this.m20 + m20, this.m21 + m21, this.m22 + m22, this.m23 + m23,
			this.m30 + m30, this.m31 + m31, this.m32 + m32, this.m33 + m33,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4d can only use 4x4 matrices")
		}

		val m00: Double; val m01: Double; val m02: Double; val m03: Double;
		val m10: Double; val m11: Double; val m12: Double; val m13: Double;
		val m20: Double; val m21: Double; val m22: Double; val m23: Double;
		val m30: Double; val m31: Double; val m32: Double; val m33: Double;

		if (other is AbstractMatrix4d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02, this.m03 - m03,
			this.m10 - m10, this.m11 - m11, this.m12 - m12, this.m13 - m13,
			this.m20 - m20, this.m21 - m21, this.m22 - m22, this.m23 - m23,
			this.m30 - m30, this.m31 - m31, this.m32 - m32, this.m33 - m33,
		)
	}

	override fun plus(other: Double): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other, this.m03 + other,
			this.m10 + other, this.m11 + other, this.m12 + other, this.m13 + other,
			this.m20 + other, this.m21 + other, this.m22 + other, this.m23 + other,
			this.m30 + other, this.m31 + other, this.m32 + other, this.m33 + other,
		)
	}

	override fun minus(other: Double): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other, this.m03 - other,
			this.m10 - other, this.m11 - other, this.m12 - other, this.m13 - other,
			this.m20 - other, this.m21 - other, this.m22 - other, this.m23 - other,
			this.m30 - other, this.m31 - other, this.m32 - other, this.m33 - other,
		)
	}

	override fun times(other: Double): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other, this.m03 * other,
			this.m10 * other, this.m11 * other, this.m12 * other, this.m13 * other,
			this.m20 * other, this.m21 * other, this.m22 * other, this.m23 * other,
			this.m30 * other, this.m31 * other, this.m32 * other, this.m33 * other,
		)
	}

	override fun scale(x: Double, y: Double, z: Double, w: Double): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02, this.m03,
			this.m10, this.m11 * y, this.m12, this.m13,
			this.m20, this.m21, this.m22 * z, this.m23,
			this.m30, this.m31, this.m32, this.m33 * w,
		)
	}

	override fun translate(x: Double, y: Double, z: Double): T {
		return createOrModify(
			this.m00, this.m01, this.m02, this.m03 + x,
			this.m10, this.m11, this.m12, this.m13 + y,
			this.m20, this.m21, this.m22, this.m23 + z,
			this.m30, this.m31, this.m32, this.m33,
		)
	}

	open val translation: Vector3d
		get() {
			return Vector3d(m03, m13, m23)
		}

	override fun div(other: Double): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other, this.m03 / other,
			this.m10 / other, this.m11 / other, this.m12 / other, this.m13 / other,
			this.m20 / other, this.m21 / other, this.m22 / other, this.m23 / other,
			this.m30 / other, this.m31 / other, this.m32 / other, this.m33 / other,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 4 || other.rows != 4) {
			throw IllegalArgumentException("Concrete Matrix4d can only use 4x4 matrices")
		}

		val m00: Double; val m01: Double; val m02: Double; val m03: Double;
		val m10: Double; val m11: Double; val m12: Double; val m13: Double;
		val m20: Double; val m21: Double; val m22: Double; val m23: Double;
		val m30: Double; val m31: Double; val m32: Double; val m33: Double;

		if (other is AbstractMatrix4d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02; m03 = other.m03;
			m10 = other.m10; m11 = other.m11; m12 = other.m12; m13 = other.m13;
			m20 = other.m20; m21 = other.m21; m22 = other.m22; m23 = other.m23;
			m30 = other.m30; m31 = other.m31; m32 = other.m32; m33 = other.m33;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2]; m03 = other[0, 3];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2]; m13 = other[1, 3];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2]; m23 = other[2, 3];
			m30 = other[3, 0]; m31 = other[3, 1]; m32 = other[3, 2]; m33 = other[3, 3];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20 +
					this.m03 * m30

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20 +
					this.m13 * m30

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20 +
					this.m23 * m30

		val newm30 =
			this.m30 * m00 +
					this.m31 * m10 +
					this.m32 * m20 +
					this.m33 * m30

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21 +
					this.m03 * m31

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21 +
					this.m13 * m31

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21 +
					this.m23 * m31

		val newm31 =
			this.m30 * m01 +
					this.m31 * m11 +
					this.m32 * m21 +
					this.m33 * m31

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22 +
					this.m03 * m32

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22 +
					this.m13 * m32

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22 +
					this.m23 * m32

		val newm32 =
			this.m30 * m02 +
					this.m31 * m12 +
					this.m32 * m22 +
					this.m33 * m32

		// четвёртый столбец
		val newm03 =
			this.m00 * m03 +
					this.m01 * m13 +
					this.m02 * m23 +
					this.m03 * m33

		val newm13 =
			this.m10 * m03 +
					this.m11 * m13 +
					this.m12 * m23 +
					this.m13 * m33

		val newm23 =
			this.m20 * m03 +
					this.m21 * m13 +
					this.m22 * m23 +
					this.m23 * m33

		val newm33 =
			this.m30 * m03 +
					this.m31 * m13 +
					this.m32 * m23 +
					this.m33 * m33

		return createOrModify(
			newm00, newm01, newm02, newm03,
			newm10, newm11, newm12, newm13,
			newm20, newm21, newm22, newm23,
			newm30, newm31, newm32, newm33,
		)
	}

	override fun translateWithScale(x: Double, y: Double, z: Double): T {
		return createOrModify(
			m00, m01, m02, m03 + x * m00 + y * m01 + z * m02,
			m10, m11, m12, m13 + x * m10 + y * m11 + z * m12,
			m20, m21, m22, m23 + x * m20 + y * m21 + z * m22,
			m30, m31, m32, m33,
		)
	}

	val absoluteMatrix get() = Matrix4d(
		m00 = m00.absoluteValue, m01 = m01.absoluteValue, m02 = m02.absoluteValue, m03 = m03.absoluteValue,
		m10 = m10.absoluteValue, m11 = m11.absoluteValue, m12 = m12.absoluteValue, m13 = m13.absoluteValue,
		m20 = m20.absoluteValue, m21 = m21.absoluteValue, m22 = m22.absoluteValue, m23 = m23.absoluteValue,
		m30 = m30.absoluteValue, m31 = m31.absoluteValue, m32 = m32.absoluteValue, m33 = m33.absoluteValue,
	)
}

data class Matrix4d(
	override val m00: Double = 1.0, override val m01: Double = 0.0, override val m02: Double = 0.0, override val m03: Double = 0.0,
	override val m10: Double = 0.0, override val m11: Double = 1.0, override val m12: Double = 0.0, override val m13: Double = 0.0,
	override val m20: Double = 0.0, override val m21: Double = 0.0, override val m22: Double = 1.0, override val m23: Double = 0.0,
	override val m30: Double = 0.0, override val m31: Double = 0.0, override val m32: Double = 0.0, override val m33: Double = 1.0,
) : AbstractMatrix4d<Matrix4d>() {
	override fun createOrModify(
		m00: Double, m01: Double, m02: Double, m03: Double,
		m10: Double, m11: Double, m12: Double, m13: Double,
		m20: Double, m21: Double, m22: Double, m23: Double,
		m30: Double, m31: Double, m32: Double, m33: Double,
	): Matrix4d {
		return Matrix4d(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	fun toMutableMatrix(): MutableMatrix4d {
		return MutableMatrix4d(
			m00 = m00, m01 = m01, m02 = m02, m03 = m03,
			m10 = m10, m11 = m11, m12 = m12, m13 = m13,
			m20 = m20, m21 = m21, m22 = m22, m23 = m23,
			m30 = m30, m31 = m31, m32 = m32, m33 = m33,
		)
	}

	companion object {
		val IDENTITY = Matrix4d()

		val SCREEN_FLIP = IDENTITY.let {
			return@let it * Vector3d.FORWARD.rotateAroundThis(-PI / 2)
		}

		/**
		 * Возвращает матрицу ортографической проекции, с ИНВЕНТИРОВАННОЙ y координатой, и с добавлением 2f
		 *
		 * Т.е. любое значение компоненты вектора y будет иметь противоположный знак после перемножения на данную матрицу
		 *
		 * Смысл данного изменения знака в преобразовании экранных координат OpenGL к вменяемому виду. Многие примеры указывают Z как отрицательную компоненту,
		 * что так же "убирает" это недоумение, только вот у нас Z это глубина, а не высота
		 *
		 * y = 0 будет соответствовать верхнему левому углу окна
		 */
		fun ortho(left: Double, right: Double, bottom: Double, top: Double, zNear: Double, zFar: Double): Matrix4d {
			return Matrix4d(
				m00 = 2f / (right - left),
				m11 = -2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom) + 2f,
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Возвращает матрицу ортографической проекции, без инвентирования
		 *
		 * y = 0 будет соответствовать нижнему левому углу окна
		 */
		fun orthoDirect(left: Double, right: Double, bottom: Double, top: Double, zNear: Double, zFar: Double): Matrix4d {
			return Matrix4d(
				m00 = 2f / (right - left),
				m11 = 2f / (top - bottom),
				m22 = 2f / (zFar - zNear),
				m03 = -(right + left) / (right - left),
				m13 = -(top + bottom) / (top - bottom),
				m23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		fun perspective(fov: Double, zFar: Double, zNear: Double): Matrix4d {
			val scale = (1.0 / (tan(Math.toRadians(fov.toDouble()) / 2.0))).toDouble()
			val r = zFar - zNear

			return Matrix4d(
				m00 = scale,
				m11 = scale,
				m22 = -zFar / r,
				m23 = -1.0,
				m32 = -zFar * zNear / r,
			)
		}
	}
}

data class MutableMatrix4d(
	override var m00: Double = 1.0, override var m01: Double = 0.0, override var m02: Double = 0.0, override var m03: Double = 0.0,
	override var m10: Double = 0.0, override var m11: Double = 1.0, override var m12: Double = 0.0, override var m13: Double = 0.0,
	override var m20: Double = 0.0, override var m21: Double = 0.0, override var m22: Double = 1.0, override var m23: Double = 0.0,
	override var m30: Double = 0.0, override var m31: Double = 0.0, override var m32: Double = 0.0, override var m33: Double = 1.0,
) : AbstractMatrix4d<MutableMatrix4d>() {
	override fun createOrModify(
		m00: Double, m01: Double, m02: Double, m03: Double,
		m10: Double, m11: Double, m12: Double, m13: Double,
		m20: Double, m21: Double, m22: Double, m23: Double,
		m30: Double, m31: Double, m32: Double, m33: Double,
	): MutableMatrix4d {
		this.m00 = m00; this.m01 = m01; this.m02 = m02; this.m03 = m03
		this.m10 = m10; this.m11 = m11; this.m12 = m12; this.m13 = m13
		this.m20 = m20; this.m21 = m21; this.m22 = m22; this.m23 = m23
		this.m30 = m30; this.m31 = m31; this.m32 = m32; this.m33 = m33

		return this
	}

	override var translation: Vector3d
		get() {
			return Vector3d(m03, m13, m23)
		}
		set(value) {
			translate(value)
		}

	fun set(row: Int, column: Int, value: Double) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				3 -> m30 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				3 -> m31 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				3 -> m32 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			3 -> when (row) {
				0 -> m03 = value
				1 -> m13 = value
				2 -> m23 = value
				3 -> m33 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}

abstract class AbstractMatrix3f<T : AbstractMatrix3f<T>> : FloatMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 3 * 3).also { it.order(ByteOrder.nativeOrder()) }.asFloatBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectFloatBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toFloatBuffer(columnMajor: Boolean = true): FloatBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Float; abstract val m01: Float; abstract val m02: Float
	abstract val m10: Float; abstract val m11: Float; abstract val m12: Float
	abstract val m20: Float; abstract val m21: Float; abstract val m22: Float

	open val a11 get() = m00; open val a12 get() = m01; open val a13 get() = m02
	open val a21 get() = m10; open val a22 get() = m11; open val a23 get() = m12
	open val a31 get() = m20; open val a32 get() = m21; open val a33 get() = m22

	override val columns: Int
		get() = 3

	override val rows: Int
		get() = 3

	override fun get(row: Int, column: Int): Float {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is AbstractMatrix3f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02,
			this.m10 + m10, this.m11 + m11, this.m12 + m12,
			this.m20 + m20, this.m21 + m21, this.m22 + m22,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrices")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is AbstractMatrix3f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02,
			this.m10 - m10, this.m11 - m11, this.m12 - m12,
			this.m20 - m20, this.m21 - m21, this.m22 - m22,
		)
	}

	override fun plus(other: Float): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other,
			this.m10 + other, this.m11 + other, this.m12 + other,
			this.m20 + other, this.m21 + other, this.m22 + other,
		)
	}

	override fun minus(other: Float): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other,
			this.m10 - other, this.m11 - other, this.m12 - other,
			this.m20 - other, this.m21 - other, this.m22 - other,
		)
	}

	override fun times(other: Float): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other,
			this.m10 * other, this.m11 * other, this.m12 * other,
			this.m20 * other, this.m21 * other, this.m22 * other,
		)
	}

	override fun div(other: Float): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other,
			this.m10 / other, this.m11 / other, this.m12 / other,
			this.m20 / other, this.m21 / other, this.m22 / other,
		)
	}

	override fun scale(x: Float, y: Float, z: Float, w: Float): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02,
			this.m10, this.m11 * y, this.m12,
			this.m20, this.m21, this.m22 * z,
		)
	}

	override fun translate(x: Float, y: Float, z: Float): T {
		return createOrModify(
			this.m00, this.m01, this.m02 + x,
			this.m10, this.m11, this.m12 + y,
			this.m20, this.m21, this.m22,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is IMatrixLikeFloat) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrixes")
		}

		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (other is AbstractMatrix3f<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22

		return createOrModify(
			newm00, newm01, newm02,
			newm10, newm11, newm12,
			newm20, newm21, newm22,
		)
	}

	override fun translateWithScale(x: Float, y: Float, z: Float): T {
		return createOrModify(
			m00, m01, m02 + x * m00 + y * m01,
			m10, m11, m12 + x * m10 + y * m11,
			m20, m21, m22,
		)
	}

	val determinant get() = m00 * m11 * m22 + m01 * m12 * m20 + m02 * m10 * m21 - m02 * m11 * m20 - m00 * m12 * m21 - m01 * m10 * m22

	val absoluteMatrix get() = Matrix3f(
		m00 = m00.absoluteValue, m01 = m01.absoluteValue, m02 = m02.absoluteValue,
		m10 = m10.absoluteValue, m11 = m11.absoluteValue, m12 = m12.absoluteValue,
		m20 = m20.absoluteValue, m21 = m21.absoluteValue, m22 = m22.absoluteValue,
	)
}

data class Matrix3f(
	override val m00: Float = 1f, override val m01: Float = 0f, override val m02: Float = 0f,
	override val m10: Float = 0f, override val m11: Float = 1f, override val m12: Float = 0f,
	override val m20: Float = 0f, override val m21: Float = 0f, override val m22: Float = 1f,
) : AbstractMatrix3f<Matrix3f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): Matrix3f {
		return Matrix3f(
			m00 = m00, m01 = m01, m02 = m02,
			m10 = m10, m11 = m11, m12 = m12,
			m20 = m20, m21 = m21, m22 = m22,
		)
	}

	companion object {
		val IDENTITY = Matrix3f()
	}
}

data class MutableMatrix3f(
	override var m00: Float = 1f, override var m01: Float = 0f, override var m02: Float = 0f,
	override var m10: Float = 0f, override var m11: Float = 1f, override var m12: Float = 0f,
	override var m20: Float = 0f, override var m21: Float = 0f, override var m22: Float = 1f,
) : AbstractMatrix3f<MutableMatrix3f>() {
	override fun createOrModify(
		m00: Float, m01: Float, m02: Float,
		m10: Float, m11: Float, m12: Float,
		m20: Float, m21: Float, m22: Float,
	): MutableMatrix3f {
		this.m00 = m00; this.m01 = m01; this.m02 = m02
		this.m10 = m10; this.m11 = m11; this.m12 = m12
		this.m20 = m20; this.m21 = m21; this.m22 = m22

		return this
	}

	fun load(from: FloatMatrix<*>) {
		val m00: Float; val m01: Float; val m02: Float;
		val m10: Float; val m11: Float; val m12: Float;
		val m20: Float; val m21: Float; val m22: Float;

		if (from is AbstractMatrix3f<*>) {
			m00 = from.m00; m01 = from.m01; m02 = from.m02;
			m10 = from.m10; m11 = from.m11; m12 = from.m12;
			m20 = from.m20; m21 = from.m21; m22 = from.m22;
		} else {
			m00 = from[0, 0]; m01 = from[0, 1]; m02 = from[0, 2];
			m10 = from[1, 0]; m11 = from[1, 1]; m12 = from[1, 2];
			m20 = from[2, 0]; m21 = from[2, 1]; m22 = from[2, 2];
		}

		this.m00 = m00; this.m01 = m01; this.m02 = m02
		this.m10 = m10; this.m11 = m11; this.m12 = m12
		this.m20 = m20; this.m21 = m21; this.m22 = m22
	}

	override var a11 by this::m00; override var a12 by this::m01; override var a13 by this::m02
	override var a21 by this::m10; override var a22 by this::m11; override var a23 by this::m12
	override var a31 by this::m20; override var a32 by this::m21; override var a33 by this::m22

	operator fun set(row: Int, column: Int, value: Float) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}

abstract class AbstractMatrix3d<T : AbstractMatrix3d<T>> : DoubleMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(4 * 3 * 3).also { it.order(ByteOrder.nativeOrder()) }.asDoubleBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectDoubleBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toDoubleBuffer(columnMajor: Boolean = true): DoubleBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)
			directBuffer.put(m20)

			directBuffer.put(m01)
			directBuffer.put(m11)
			directBuffer.put(m21)

			directBuffer.put(m02)
			directBuffer.put(m12)
			directBuffer.put(m22)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)
			directBuffer.put(m02)

			directBuffer.put(m10)
			directBuffer.put(m11)
			directBuffer.put(m12)

			directBuffer.put(m20)
			directBuffer.put(m21)
			directBuffer.put(m22)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Double; abstract val m01: Double; abstract val m02: Double
	abstract val m10: Double; abstract val m11: Double; abstract val m12: Double
	abstract val m20: Double; abstract val m21: Double; abstract val m22: Double

	open val a11 get() = m00; open val a12 get() = m01; open val a13 get() = m02
	open val a21 get() = m10; open val a22 get() = m11; open val a23 get() = m12
	open val a31 get() = m20; open val a32 get() = m21; open val a33 get() = m22

	override val columns: Int
		get() = 3

	override val rows: Int
		get() = 3

	override fun get(row: Int, column: Int): Double {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				2 -> m20
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				2 -> m21
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02
				1 -> m12
				2 -> m22
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Double, m01: Double, m02: Double,
		m10: Double, m11: Double, m12: Double,
		m20: Double, m21: Double, m22: Double,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3d can only use 3x3 matrices")
		}

		val m00: Double; val m01: Double; val m02: Double;
		val m10: Double; val m11: Double; val m12: Double;
		val m20: Double; val m21: Double; val m22: Double;

		if (other is AbstractMatrix3d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01, this.m02 + m02,
			this.m10 + m10, this.m11 + m11, this.m12 + m12,
			this.m20 + m20, this.m21 + m21, this.m22 + m22,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3d can only use 3x3 matrices")
		}

		val m00: Double; val m01: Double; val m02: Double;
		val m10: Double; val m11: Double; val m12: Double;
		val m20: Double; val m21: Double; val m22: Double;

		if (other is AbstractMatrix3d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01, this.m02 - m02,
			this.m10 - m10, this.m11 - m11, this.m12 - m12,
			this.m20 - m20, this.m21 - m21, this.m22 - m22,
		)
	}

	override fun plus(other: Double): T {
		return createOrModify(
			this.m00 + other, this.m01 + other, this.m02 + other,
			this.m10 + other, this.m11 + other, this.m12 + other,
			this.m20 + other, this.m21 + other, this.m22 + other,
		)
	}

	override fun minus(other: Double): T {
		return createOrModify(
			this.m00 - other, this.m01 - other, this.m02 - other,
			this.m10 - other, this.m11 - other, this.m12 - other,
			this.m20 - other, this.m21 - other, this.m22 - other,
		)
	}

	override fun times(other: Double): T {
		return createOrModify(
			this.m00 * other, this.m01 * other, this.m02 * other,
			this.m10 * other, this.m11 * other, this.m12 * other,
			this.m20 * other, this.m21 * other, this.m22 * other,
		)
	}

	override fun div(other: Double): T {
		return createOrModify(
			this.m00 / other, this.m01 / other, this.m02 / other,
			this.m10 / other, this.m11 / other, this.m12 / other,
			this.m20 / other, this.m21 / other, this.m22 / other,
		)
	}

	override fun scale(x: Double, y: Double, z: Double, w: Double): T {
		return createOrModify(
			this.m00 * x, this.m01, this.m02,
			this.m10, this.m11 * y, this.m12,
			this.m20, this.m21, this.m22 * z,
		)
	}

	override fun translate(x: Double, y: Double, z: Double): T {
		return createOrModify(
			this.m00, this.m01, this.m02 + x,
			this.m10, this.m11, this.m12 + y,
			this.m20, this.m21, this.m22,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3d can only use 3x3 matrixes")
		}

		val m00: Double; val m01: Double; val m02: Double;
		val m10: Double; val m11: Double; val m12: Double;
		val m20: Double; val m21: Double; val m22: Double;

		if (other is AbstractMatrix3d<*>) {
			m00 = other.m00; m01 = other.m01; m02 = other.m02;
			m10 = other.m10; m11 = other.m11; m12 = other.m12;
			m20 = other.m20; m21 = other.m21; m22 = other.m22;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1]; m02 = other[0, 2];
			m10 = other[1, 0]; m11 = other[1, 1]; m12 = other[1, 2];
			m20 = other[2, 0]; m21 = other[2, 1]; m22 = other[2, 2];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10 +
					this.m02 * m20

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10 +
					this.m12 * m20

		val newm20 =
			this.m20 * m00 +
					this.m21 * m10 +
					this.m22 * m20

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11 +
					this.m02 * m21

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11 +
					this.m12 * m21

		val newm21 =
			this.m20 * m01 +
					this.m21 * m11 +
					this.m22 * m21

		// третий столбец
		val newm02 =
			this.m00 * m02 +
					this.m01 * m12 +
					this.m02 * m22

		val newm12 =
			this.m10 * m02 +
					this.m11 * m12 +
					this.m12 * m22

		val newm22 =
			this.m20 * m02 +
					this.m21 * m12 +
					this.m22 * m22

		return createOrModify(
			newm00, newm01, newm02,
			newm10, newm11, newm12,
			newm20, newm21, newm22,
		)
	}

	override fun translateWithScale(x: Double, y: Double, z: Double): T {
		return createOrModify(
			m00, m01, m02 + x * m00 + y * m01,
			m10, m11, m12 + x * m10 + y * m11,
			m20, m21, m22,
		)
	}

	val determinant get() = m00 * m11 * m22 + m01 * m12 * m20 + m02 * m10 * m21 - m02 * m11 * m20 - m00 * m12 * m21 - m01 * m10 * m22

	val absoluteMatrix get() = Matrix3d(
		m00 = m00.absoluteValue, m01 = m01.absoluteValue, m02 = m02.absoluteValue,
		m10 = m10.absoluteValue, m11 = m11.absoluteValue, m12 = m12.absoluteValue,
		m20 = m20.absoluteValue, m21 = m21.absoluteValue, m22 = m22.absoluteValue,
	)

	/// Solve A * x = b, where b is a column vector. This is more efficient
	/// than computing the inverse in one-shot cases.
	fun solve(b: IVector3d<*>): Vector3d {
		var determinant = determinant

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		}

		return Vector3d(
			x = determinant * scalarDotWithCross(
				b.x, b.y, b.z,
				m01, m11, m21,
				m02, m12, m22,
			),

			y = determinant * scalarDotWithCross(
				m00, m10, m20,
				b.x, b.y, b.z,
				m02, m12, m22,
			),

			z = determinant * scalarDotWithCross(
				m00, m10, m20,
				m01, m11, m21,
				b.x, b.y, b.z,
			),
		)
	}

	/// Solve A * x = b, where b is a column vector. This is more efficient
	/// than computing the inverse in one-shot cases.
	fun solve(b: IVector2d<*>): Vector2d {
		var determinant = determinant

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		}

		return Vector2d(
			x = determinant * (m11 * b.x - m01 * b.y),
			y = determinant * (m00 * b.y - m10 * b.x),
		)
	}

	fun getInverse2(): Matrix3d {
		val a = m00
		val b = m01
		val c = m10
		val d = m11

		var determinant = a * d - b * c

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		}

		return Matrix3d(
			m00 = determinant * d, m01 = -determinant * b, m02 = 0.0,
			m10 = -determinant * c, m11 = determinant * a, m12 = 0.0,
			m20 = 0.0, m21 = 0.0, m22 = 0.0,
		)
	}

	fun getInverse(): Matrix3d {
		var determinant = determinant

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		} else {
			return Matrix3d(
				0.0, 0.0, 0.0,
				0.0, 0.0, 0.0,
				0.0, 0.0, 0.0,
			)
		}

		val m00 = determinant * (a22 * a33 - a23 * a23)
		val m10 = determinant * (a13 * a23 - a12 * a33)
		val m20 = determinant * (a12 * a23 - a13 * a22)

		val m01 = m10
		val m11 = determinant * (a11 * a33 - a13 * a13)
		val m21 = determinant * (a13 * a12 - a11 * a23)

		val m02 = m20
		val m12 = m21
		val m22 = determinant * (a11 * a22 - a12 * a12)

		return Matrix3d(
			m00 = m00,
			m10 = m10,
			m20 = m20,
			m01 = m01,
			m11 = m11,
			m21 = m21,
			m02 = m02,
			m12 = m12,
			m22 = m22,
		)
	}
}

data class Matrix3d(
	override val m00: Double = 1.0, override val m01: Double = 0.0, override val m02: Double = 0.0,
	override val m10: Double = 0.0, override val m11: Double = 1.0, override val m12: Double = 0.0,
	override val m20: Double = 0.0, override val m21: Double = 0.0, override val m22: Double = 1.0,
) : AbstractMatrix3d<Matrix3d>() {
	override fun createOrModify(
		m00: Double, m01: Double, m02: Double,
		m10: Double, m11: Double, m12: Double,
		m20: Double, m21: Double, m22: Double,
	): Matrix3d {
		return Matrix3d(
			m00 = m00, m01 = m01, m02 = m02,
			m10 = m10, m11 = m11, m12 = m12,
			m20 = m20, m21 = m21, m22 = m22,
		)
	}

	fun asMutableMatrix(): MutableMatrix3d {
		return MutableMatrix3d(
			m00 = m00, m01 = m01, m02 = m02,
			m10 = m10, m11 = m11, m12 = m12,
			m20 = m20, m21 = m21, m22 = m22,
		)
	}

	companion object {
		val IDENTITY = Matrix3d()
	}
}

data class MutableMatrix3d(
	override var m00: Double = 1.0, override var m01: Double = 0.0, override var m02: Double = 0.0,
	override var m10: Double = 0.0, override var m11: Double = 1.0, override var m12: Double = 0.0,
	override var m20: Double = 0.0, override var m21: Double = 0.0, override var m22: Double = 1.0,
) : AbstractMatrix3d<MutableMatrix3d>() {
	override fun createOrModify(
		m00: Double, m01: Double, m02: Double,
		m10: Double, m11: Double, m12: Double,
		m20: Double, m21: Double, m22: Double,
	): MutableMatrix3d {
		this.m00 = m00; this.m01 = m01; this.m02 = m02
		this.m10 = m10; this.m11 = m11; this.m12 = m12
		this.m20 = m20; this.m21 = m21; this.m22 = m22

		return this
	}

	override var a11 by this::m00; override var a12 by this::m01; override var a13 by this::m02
	override var a21 by this::m10; override var a22 by this::m11; override var a23 by this::m12
	override var a31 by this::m20; override var a32 by this::m21; override var a33 by this::m22

	fun load(from: DoubleMatrix<*>) {
		val m00: Double; val m01: Double; val m02: Double;
		val m10: Double; val m11: Double; val m12: Double;
		val m20: Double; val m21: Double; val m22: Double;

		if (from is AbstractMatrix3d<*>) {
			m00 = from.m00; m01 = from.m01; m02 = from.m02;
			m10 = from.m10; m11 = from.m11; m12 = from.m12;
			m20 = from.m20; m21 = from.m21; m22 = from.m22;
		} else {
			m00 = from[0, 0]; m01 = from[0, 1]; m02 = from[0, 2];
			m10 = from[1, 0]; m11 = from[1, 1]; m12 = from[1, 2];
			m20 = from[2, 0]; m21 = from[2, 1]; m22 = from[2, 2];
		}

		this.m00 = m00; this.m01 = m01; this.m02 = m02
		this.m10 = m10; this.m11 = m11; this.m12 = m12
		this.m20 = m20; this.m21 = m21; this.m22 = m22
	}

	fun setColumns(column1: IVector3d<*>, column2: IVector3d<*>, column3: IVector3d<*>) {
		m00 = column1.x
		m10 = column1.y
		m20 = column1.z

		m01 = column2.x
		m11 = column2.y
		m21 = column2.y

		m02 = column3.x
		m12 = column3.y
		m22 = column3.y
	}

	fun setIdentity() {
		m00 = 1.0
		m01 = 0.0
		m02 = 0.0

		m10 = 0.0
		m11 = 1.0
		m12 = 0.0

		m20 = 0.0
		m21 = 0.0
		m22 = 1.0
	}

	fun zero() {
		m00 = 0.0
		m01 = 0.0
		m02 = 0.0

		m10 = 0.0
		m11 = 0.0
		m12 = 0.0

		m20 = 0.0
		m21 = 0.0
		m22 = 0.0
	}

	operator fun set(row: Int, column: Int, value: Double) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				2 -> m20 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				2 -> m21 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			2 -> when (row) {
				0 -> m02 = value
				1 -> m12 = value
				2 -> m22 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}

abstract class AbstractMatrix2d<T : AbstractMatrix2d<T>> : DoubleMatrix<T> {
	companion object {
		private val directBuffer by lazy { ByteBuffer.allocateDirect(8 * 2 * 2).also { it.order(ByteOrder.nativeOrder()) }.asDoubleBuffer()!! }
	}

	/**
	 * Наполняет синглтон и возвращает DirectFloatBuffer который необходимо использовать сразу,
	 * нет никаких гарантий что он не будет изменён в дальнейшем
	 *
	 * Функция предназначена исключительно для загрузки данной матрицы в память видеокарты
	 *
	 * Данный буфер всегда будет одним и тем же, но его содержимое обновлено
	 */
	fun toDoubleBuffer(columnMajor: Boolean = true): DoubleBuffer {
		val directBuffer = directBuffer
		directBuffer.position(0)

		if (columnMajor) {
			directBuffer.put(m00)
			directBuffer.put(m10)

			directBuffer.put(m01)
			directBuffer.put(m11)
		} else {
			directBuffer.put(m00)
			directBuffer.put(m01)

			directBuffer.put(m10)
			directBuffer.put(m11)
		}

		directBuffer.position(0)
		return directBuffer
	}

	abstract val m00: Double; abstract val m01: Double;
	abstract val m10: Double; abstract val m11: Double;

	override val columns: Int
		get() = 2

	override val rows: Int
		get() = 2

	override fun get(row: Int, column: Int): Double {
		return when (column) {
			0 -> when (row) {
				0 -> m00
				1 -> m10
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01
				1 -> m11
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}

	protected abstract fun createOrModify(
		m00: Double, m01: Double,
		m10: Double, m11: Double,
	): T

	override fun plus(other: IMatrix): T {
		if (other !is DoubleMatrix<*>) {
			throw IllegalArgumentException("Can not use $other for addition")
		}

		if (other.columns != 2 || other.rows != 2) {
			throw IllegalArgumentException("Concrete Matrix2f can only use 2x2 matrices")
		}

		val m00: Double; val m01: Double;
		val m10: Double; val m11: Double;

		if (other is AbstractMatrix2d<*>) {
			m00 = other.m00; m01 = other.m01;
			m10 = other.m10; m11 = other.m11;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1];
			m10 = other[1, 0]; m11 = other[1, 1];
		}

		return createOrModify(
			this.m00 + m00, this.m01 + m01,
			this.m10 + m10, this.m11 + m11,
		)
	}

	override fun minus(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for subtraction")
		}

		if (other.columns != 2 || other.rows != 2) {
			throw IllegalArgumentException("Concrete Matrix2f can only use 2x2 matrices")
		}

		val m00: Double; val m01: Double;
		val m10: Double; val m11: Double;

		if (other is AbstractMatrix2d<*>) {
			m00 = other.m00; m01 = other.m01;
			m10 = other.m10; m11 = other.m11;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1];
			m10 = other[1, 0]; m11 = other[1, 1];
		}

		return createOrModify(
			this.m00 - m00, this.m01 - m01,
			this.m10 - m10, this.m11 - m11,
		)
	}

	override fun plus(other: Double): T {
		return createOrModify(
			this.m00 + other, this.m01 + other,
			this.m10 + other, this.m11 + other,
		)
	}

	override fun minus(other: Double): T {
		return createOrModify(
			this.m00 - other, this.m01 - other,
			this.m10 - other, this.m11 - other,
		)
	}

	override fun times(other: Double): T {
		return createOrModify(
			this.m00 * other, this.m01 * other,
			this.m10 * other, this.m11 * other,
		)
	}

	override fun div(other: Double): T {
		return createOrModify(
			this.m00 / other, this.m01 / other,
			this.m10 / other, this.m11 / other,
		)
	}

	override fun scale(x: Double, y: Double, z: Double, w: Double): T {
		return createOrModify(
			this.m00 * x, this.m01,
			this.m10, this.m11 * y,
		)
	}

	override fun translate(x: Double, y: Double, z: Double): T {
		return createOrModify(
			this.m00, this.m01 + x,
			this.m10, this.m11,
		)
	}

	override fun times(other: IMatrix): T {
		if (other !is IMatrixLikeDouble) {
			throw IllegalArgumentException("Can not use $other for multiplication")
		}

		if (other.columns != 3 || other.rows != 3) {
			throw IllegalArgumentException("Concrete Matrix3f can only use 3x3 matrixes")
		}

		val m00: Double; val m01: Double;
		val m10: Double; val m11: Double;

		if (other is AbstractMatrix2d<*>) {
			m00 = other.m00; m01 = other.m01;
			m10 = other.m10; m11 = other.m11;
		} else {
			m00 = other[0, 0]; m01 = other[0, 1];
			m10 = other[1, 0]; m11 = other[1, 1];
		}

		// первый столбец
		val newm00 =
			this.m00 * m00 +
					this.m01 * m10

		val newm10 =
			this.m10 * m00 +
					this.m11 * m10

		// второй столбец
		val newm01 =
			this.m00 * m01 +
					this.m01 * m11

		val newm11 =
			this.m10 * m01 +
					this.m11 * m11

		return createOrModify(
			newm00, newm01,
			newm10, newm11,
		)
	}

	override fun translateWithScale(x: Double, y: Double, z: Double): T {
		return createOrModify(
			m00, m01 + x * m00,
			m10, m11,
		)
	}

	val determinant: Double get() = m00 * m11 - m01 * m10

	val absoluteMatrix get() = Matrix2d(
		m00 = m00.absoluteValue, m01 = m01.absoluteValue,
		m10 = m10.absoluteValue, m11 = m11.absoluteValue,
	)

	fun getInverse(): Matrix2d {
		val a = m00
		val b = m01
		val c = m10
		val d = m11

		var determinant = determinant

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		}

		return Matrix2d(
			m00 = determinant * d, m01 = -determinant * b,
			m10 = -determinant * c, m11 = determinant * a,
		)
	}

	/// Solve A * x = b, where b is a column vector. This is more efficient
	/// than computing the inverse in one-shot cases.
	fun solve(b: IVector2d<*>): Vector2d {
		var determinant = determinant

		if (determinant != 0.0) {
			determinant = 1.0 / determinant
		}

		return Vector2d(
			x = determinant * (m11 * b.x - m01 * b.y),
			y = determinant * (m00 * b.y - m10 * b.x),
		)
	}
}

// ex.x = m00
// ex.y = m10
// ey.x = m01
// ey.y = m11
data class Matrix2d(
	override val m00: Double = 1.0, override val m01: Double = 0.0,
	override val m10: Double = 0.0, override val m11: Double = 1.0,
) : AbstractMatrix2d<Matrix2d>() {
	override fun createOrModify(
		m00: Double, m01: Double,
		m10: Double, m11: Double,
	): Matrix2d {
		return Matrix2d(
			m00 = m00, m01 = m01,
			m10 = m10, m11 = m11,
		)
	}

	fun asMutableMatrix(): MutableMatrix2d {
		return MutableMatrix2d(
			m00 = m00, m01 = m01,
			m10 = m10, m11 = m11,
		)
	}

	companion object {
		val IDENTITY = Matrix2d()
	}
}

data class MutableMatrix2d(
	override var m00: Double = 1.0, override var m01: Double = 0.0,
	override var m10: Double = 0.0, override var m11: Double = 1.0,
) : AbstractMatrix2d<MutableMatrix2d>() {
	override fun createOrModify(
		m00: Double, m01: Double,
		m10: Double, m11: Double,
	): MutableMatrix2d {
		this.m00 = m00; this.m01 = m01;
		this.m10 = m10; this.m11 = m11;

		return this
	}

	fun setColumns(column1: IVector2d<*>, column2: IVector2d<*>) {
		m00 = column1.x
		m10 = column1.y

		m01 = column2.x
		m11 = column2.y
	}

	fun setIdentity() {
		m00 = 1.0
		m01 = 0.0
		m11 = 1.0
		m10 = 0.0
	}

	fun zero() {
		m00 = 0.0
		m01 = 0.0
		m11 = 0.0
		m10 = 0.0
	}

	operator fun set(row: Int, column: Int, value: Double) {
		when (column) {
			0 -> when (row) {
				0 -> m00 = value
				1 -> m10 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			1 -> when (row) {
				0 -> m01 = value
				1 -> m11 = value
				else -> throw IndexOutOfBoundsException("Row: $row")
			}

			else -> throw IndexOutOfBoundsException("Column: $column")
		}
	}
}
